#include "common.h"
#include "xmodem.h"
#include "dataflash.h"

void init(void)
{
	// enable peripheral clock for PIO controller
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA) | (1<<AT91C_ID_PIOB) | (1<<AT91C_ID_PIOC);
	LED_GP_INIT;
	debug_init_default();
}

static TXmodemBlock block;

static uint8_t xmodemByteNumber=XMODEM_PAYLOAD_LENGTH;

//reads next byte downloaded via xmodem protocol
//returns 1 if the byte is last one
uint8_t getNextByteFromXmodemBlock(uint8_t *theByte)
{
	static uint8_t lastByte;
	
	if(lastByte) return(lastByte);
	
	if(xmodemByteNumber < XMODEM_PAYLOAD_LENGTH)
	{
		*theByte = block.payload[xmodemByteNumber];
		xmodemByteNumber++;
	}
	else
	{
		xmodemByteNumber=0;
		if(xmodemDownloadBlock((TXmodemBlock*)&block)==2) lastByte=1;
		*theByte = block.payload[xmodemByteNumber];
		xmodemByteNumber++;
	}
	
	
	return(lastByte);
}


void (*addressToStart)();

static void delay(void)
{
	uint32_t i = 0x200000;
	while(i--) asm volatile("nop"::);
}


int main( void )
{
	init();
	
	dataflashInit();
	
	while(1)
	{
		delay();
		debug_msg("DataFlash bootloader. Send '*' and then 'C' to continue...");
		if(debug_inkey()=='*')
		{
			delay();
			debug_msg("now send 'C'...");
			uint32_t timeout = 10;
			while(timeout--)
			{
				delay();
				if(debug_inkey()=='C') goto startProgram;
			}
			debug_msg("try again!");
		}
	}
	
	startProgram:
	
	debug_msg("Waiting for DataFlash enable jumper...");
	
	uint8_t dfStatus;
	do
	{
		dfStatus = dataflashReadStatusRegister();
	}while((dfStatus&0xC0) != 0x80);
	
	debug_txt("ok");
	
	LED_GP_ON;
	
	debug_msg("press a key:");
	debug_msg(" * B to upload a new bootloader");
	debug_msg(" * any other to upload a new application");
	
	uint8_t key = debug_waitkey();
	
	const uint8_t APPLICATION_FIRST_BLOCK = 0x0A;
	
	uint16_t offset;
	
	if((key=='b')||(key=='B'))
	{
		offset = 0;
		debug_msg("loading bootloader...");
	}
	else
	{
		offset = APPLICATION_FIRST_BLOCK;
		debug_msg("loading application...");
	}
	
	const uint16_t DF_BUFFER_SIZE = 528;
	uint32_t dataflashActualBuffer[132];
	uint8_t *dataflashBuffer;
	dataflashBuffer = (uint8_t*)dataflashActualBuffer;
	
	uint16_t i;
	uint32_t dfSector = 0;
	uint8_t lastByte = 0;
	
	do
	{
		//obtain block from pc
		for(i=0;i<DF_BUFFER_SIZE;i++)
		{
			uint8_t nextByte = 0;
			if(getNextByteFromXmodemBlock(&nextByte) == 1) lastByte = 1;
			dataflashBuffer[i] = nextByte;
		}
		
		//write block to dataflash
		dataflashWriteBlock(dfSector+offset, dataflashBuffer);
		
		dfSector++;
		
	}while(lastByte==0);	//finish the loop after last write
	
	debug_value("DF secotrs written: ",dfSector);
	
	//read back the first sector from dataflash
	
	//initialize read from dataflash
	dataflashStartNewStream();
	
	if(offset)
	{
		for(i=0;i<(DF_BUFFER_SIZE*offset);i++)
		{
			//dummy reads
			dataflashReadStream();
		}
	}
	
	//read the first block from dataflash
	for(i=0;i<DF_BUFFER_SIZE;i++)
	{
		dataflashBuffer[i] = dataflashReadStream();
	}
	
	//terminate read from dataflash
	dataflashEndReadStream();
	
	//replace 6th (reserved) ARM vector with data length
	dataflashActualBuffer[5] = 4092;
	
	if(offset==0)
	{
		//up to 4KB of program code for boot region
		//if(dataflashActualBuffer[5] > 4095) dataflashActualBuffer[5] = 4095;
	}
	else
	{
		//up to 1MB of program code
		if(dataflashActualBuffer[5] > 0x100000) dataflashActualBuffer[5] = 0x100000;
	}
	
	dataflashWriteBlock(offset, dataflashBuffer);
	
	debug_msg("DataFlash contents:\r\n");
	for(i=0;i<8;i++)
	{
		debug32_t(dataflashActualBuffer[i]);
		debug_chr(' ');
	}
	//debug_array(dataflashBuffer,0x20);
	
	LED_GP_OFF;
	
	dataflashDeactivate();
	
	debug_msg("Ready!      ");
	
	AT91C_BASE_RSTC->RSTC_RCR = (0xA5 << 24) | AT91C_RSTC_PERRST | AT91C_RSTC_PROCRST;
	
	return(0);
}

