#include "common.h"
#include "dataflash.h"

//HARDWARE DEPENDENT
#define DF_CS_MASK			AT91C_PIO_PC11	//NPCS1 controlled by PIO (simple)
#define DF_CS_PIO			AT91C_BASE_PIOC	//NPCS1 controlled by PIO (simple)
#define DATAFLASH_NPCS		1	//NPCS register number (0...3)
#define DATAFLASH_NPCS_SEL	1	//code to activate NPCS pin: 0 for NPCS0, 1 for NPCS1, 3 for NPCS2, 7 for NPCS3
#define DATAFLASH_SPI		AT91C_BASE_SPI0	//base address of spi used for dataflash interface
#define SPI_PIO				AT91C_BASE_PIOA
#define SPI_MASK			( AT91C_PA0_SPI0_MISO | AT91C_PA1_SPI0_MOSI | AT91C_PA2_SPI0_SPCK )
#define SPI_PERIPH			PIO_ASR

#define DF_CS_INIT			{DF_CS_PIO->PIO_PER = DF_CS_MASK; DF_CS_PIO->PIO_SODR = DF_CS_MASK; DF_CS_PIO->PIO_OER = DF_CS_MASK;}
#define DF_CS_0				DF_CS_PIO->PIO_CODR = DF_CS_MASK
#define DF_CS_1				DF_CS_PIO->PIO_SODR = DF_CS_MASK

#define DF_MISO_INIT		SPI_PIO->PIO_PDR = SPI_MASK; SPI_PIO->PIO_ASR = SPI_MASK
#define DF_MOSI_INIT		
#define SCK_INIT			

#define BIG_BLOCK_SIZE		528
#define SMALL_BLOCK_SIZE	264

#define READ_ONLY			0

void dataflashDeactivate(void)
{
	DF_CS_PIO->PIO_PER = DF_CS_MASK;
	DF_CS_PIO->PIO_ODR = DF_CS_MASK;
	DF_CS_PIO->PIO_PPUER = DF_CS_MASK;
}


//HARDWARE DEPENDENT
uint8_t dataflashSpiTransfer(uint8_t input)
{
    uint8_t RXedData;
	while( !( DATAFLASH_SPI->SPI_SR & AT91C_SPI_TDRE ) ); // transfer compl. wait
	DATAFLASH_SPI->SPI_TDR = ( (uint16_t)(input) | ((uint32_t)(DATAFLASH_NPCS_SEL)<<16) );
	while( !( DATAFLASH_SPI->SPI_SR & AT91C_SPI_RDRF ) ); // wait for char
	RXedData = (uint8_t)( DATAFLASH_SPI->SPI_RDR );
	return RXedData;
}


//HARDWARE DEPENDENT
void dataflashInitializeSpi(void)
{
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_SPI0);
    //spi enable and reset
    DATAFLASH_SPI->SPI_CR = AT91C_SPI_SPIEN | AT91C_SPI_SWRST;
    //SPI Mode Register: Fault detection disable, variable periph. chip select
    DATAFLASH_SPI->SPI_MR  = AT91C_SPI_MSTR | AT91C_SPI_MODFDIS | AT91C_SPI_PS;
    //enable SPI
    DATAFLASH_SPI->SPI_CR = AT91C_SPI_SPIEN;
	DF_CS_INIT;
	DF_MISO_INIT;
	DF_MOSI_INIT;
	SCK_INIT;
	
	const uint32_t DLBCT_MASK = 0x1A << 24;
	const uint32_t DLBS_MASK = 0x1A << 16;
	const uint32_t SCBR_MASK = 0x80 << 8;
	DATAFLASH_SPI->SPI_CSR[DATAFLASH_NPCS] = DLBCT_MASK | DLBS_MASK | SCBR_MASK | AT91C_SPI_CPOL;
}

//=================================================================================================
//						hardware independent part
//=================================================================================================

static volatile uint8_t newStream;

uint8_t dataflashBusyWait(void);

//sets flag which reinitializes stream access to dataflash
void dataflashStartNewStream(void)
{
	newStream = 1;
}

#if READ_ONLY == 0
static uint32_t dfWriteBlockCore(uint32_t lba, uint8_t *buf, uint16_t blockSize)
{
	uint32_t i;
	
	//write data to the first RAM buffer in dataflash chip
	DF_CS_0;       // enable DataFlash
		dataflashSpiTransfer(0x84);	//opcode: write buffer 1
		dataflashSpiTransfer(0);	// don't cares
		dataflashSpiTransfer(0);	// don't cares plus first two bits of buffer address
		dataflashSpiTransfer(0);	// buffer address (max. 2^8 = 256 pages)
		for(i=0;i<blockSize;i++)
		{
			dataflashSpiTransfer(*buf);                   // write data into SPI Data Register
			buf++;
		}
	DF_CS_1;        // disable DataFlash
	
	/*asm volatile ("nop"::);
	asm volatile ("nop"::);
	asm volatile ("nop"::);*/
	
	if(dataflashBusyWait()!=1)
	{
		//debug_msg("bsy ret 0");
		return(0);
	}
	
	
	uint8_t addrHigh;
	uint8_t addrLow;
	
	if(blockSize == BIG_BLOCK_SIZE)
	{
		addrHigh = lba >> 6;
		addrLow = lba << 2;
	}
	else
	{
		addrHigh = lba >> 7;
		addrLow = lba << 1;
	}
	
	//copy previously written RAM buffer to Flash
	//lba is dataflash page number
	DF_CS_0;		// enable DataFlash
		dataflashSpiTransfer(0x83);	// write data from buffer1 to page 
		dataflashSpiTransfer(addrHigh);
		dataflashSpiTransfer(addrLow);
		dataflashSpiTransfer(0x00);		// don't cares
	DF_CS_1;		// disable DataFlash
	
	
	
	/*asm volatile ("nop"::);
	asm volatile ("nop"::);
	asm volatile ("nop"::);*/
	
	if(dataflashBusyWait()!=1)
	{
		//debug_msg("bsy ret 0");
		return(0);
	}
	
	return(1);
}
#endif

#if READ_ONLY==0
//writes block of data into dataflash chip
//requires: lba - page to write, *buf - pointer to buffer that contains data
//returns 0 on fail (timeout) and 1 on success
uint8_t dataflashWriteBlock(uint32_t lba, uint8_t *buf)
{
	if(dataflashBusyWait()!=1)
	{
		return(0);
	}
	
	uint8_t statusReg = dataflashReadStatusRegister();
	
	switch(statusReg)
	{
		case 0xAC:	//AT45DB161D
		{
			if(dfWriteBlockCore(lba,buf,BIG_BLOCK_SIZE)==0) return(0);
			break;
		}
		default:
/*		case 0xA4:	//AT45DB081
		case 0x9C:	//AT45DB041*/
		{
			lba = lba * 2;
			if(dfWriteBlockCore(lba,buf,SMALL_BLOCK_SIZE)==0) return(0);
			buf+=SMALL_BLOCK_SIZE;
			lba++;
			if(dfWriteBlockCore(lba,buf,SMALL_BLOCK_SIZE)==0) return(0);
			break;
		}
		//default:
			//return(0);
	}
	
	return(1);
}
#endif

//read next byte from dataflash
//sequential read should be initialized first with call to dataflashStartNewStream
//returns byte read from dataflash chip
uint8_t dataflashReadStream(void)
{
	static uint32_t byteCounter;
	if(newStream)
	{
		byteCounter = 0;
		dataflashBusyWait();
		DF_CS_0; //activate dataflash
		dataflashSpiTransfer(0xE8);	//write opcode
		dataflashSpiTransfer(0);	//address byte
		dataflashSpiTransfer(0);	//address byte
		dataflashSpiTransfer(0);	//address byte
		dataflashSpiTransfer(0);	//don't care byte
		dataflashSpiTransfer(0);	//don't care byte
		dataflashSpiTransfer(0);	//don't care byte
		dataflashSpiTransfer(0);	//don't care byte
		newStream = 0;
	}
	
	return(dataflashSpiTransfer(0xFF));
}


//terminate stream access to dataflash
void dataflashEndReadStream(void)
{
	DF_CS_1;	//deactivate dataflash
}


//waits for bit 7 set in dataflash status register
//returns 0 if no timeout occured
//returns 1 on timeout
uint8_t dataflashBusyWait(void)
{
	uint32_t retries = 0xFFFF;
	while(retries--)
	{
		if(dataflashReadStatusRegister() & 0x80) return(1);
	}
	return(0);
}


//returns contents of dataflash status register
uint8_t dataflashReadStatusRegister(void)
{
	uint8_t statusRegister;
	DF_CS_0;
	dataflashSpiTransfer(0xD7);
	statusRegister = dataflashSpiTransfer(0xFF);
	DF_CS_1;
	return(statusRegister);
}


//initializes pins and interface that access dataflash memory chip
void dataflashInit(void)
{
	dataflashInitializeSpi();
}




