#include "common.h"
#include "fs_tools.h"


static FATFS fs;				/* File system object */
static FIL fileR;				/* File object */
static FIL fileW;				/* File object */
static FILINFO fileInfo;		// file/dir info object

//internal funcs
FRESULT fsSetSize(FIL *fp, uint32_t size);
FRESULT fsDecreaseSize(FIL *fp, uint32_t bytes);
FRESULT fsFindValueForward(FIL *fp, uint8_t byte);


//=================================================================================================
//								HIGH-LEVEL UTILS (USE FILE NAME)
//=================================================================================================

//simple util to convert integer (param i) into ASCII string stored in buf
//converts values from 0 to 999 into "000" to "999" strings
//returns nonzero if conversion was OK, else returns 0
static uint8_t itoa(char *buf, uint16_t i)
{
	if(i > 999) return FAIL;
	buf[0] = i/100 + '0';
	buf[1] = (i%100)/10 + '0';
	buf[2] = i%10 + '0';
	buf[3] = 0;
	return(OK);
}


//base: pointer to base of file name. Must be max 5 chars long
//ext: pointer to extension of file name. Must be max 3 chars long
//outputName: fileName of non-existing file
//returns OK if free file name was found or FAIL if not
uint8_t fsFindFreeFileName(const char *base, const char *ext, char *outputName)
{
	FILINFO info;
	uint16_t nameCounter = 0;
	FRESULT res;
	
	do
	{
		//prepare file name to test
		
		//get length of base
		uint8_t len = strlen(base);
		if(len > 5) return(FAIL);
		
		//copy base to file name to be tested
		strcpy(outputName,base);
		
		//convert file number into string
		itoa(&outputName[len],nameCounter);
		
		//dot before extension
		strcat(outputName,".");
		
		//add extension to file name
		strcat(outputName,ext);
		
		//display name to be tested
		debug_msg("trying name: "); debug_txt(outputName);
		
		//test the new name
		res = f_stat((const char*)outputName,&info);
		
		//increment file number to be tested
		nameCounter++;
		if(nameCounter > 999) {debug_msg("fsFindFreeFileName: FAIL, couldn't find any free file name"); return(FAIL);}
		
	}while(res == FR_OK);	//exit if file was not found (doesn't exist)
	
	//display file name to be used (option)
	//debug_msg("found free name: "); debug_txt(outputName);
	
	return(OK);
}





/*lists all files that are in root directory or in a subdirectory
	...useful in players, viewers etc.
	
	parameters:
	
	buffer is a pointer to 2-dim array that holds list of files/dirs,
		for example array[x][y], stores x entries of length y
	pathLen is the max length for 1 entry (file name with path)
	maxItems is the max number of entries to obtain (max. file list length)
	extFilter is a pointer to string that allows filtering extensions:
		"*": lists all withing directory (dirName)
		"" (NULL, empty string): lists only directories (subdirectories) in current directory
		any other string with 3 chars or less is treated as an extension,
			i.e."MP3" will list files with .MP3 extension only.
		extension filter and file name are converted to lower case before comparing
	
	return value:
	
	function returns number of entries written to buffer (typically number of files/dirs) in specified directory
*/
uint16_t fsListAllFiles(const char *dirName, char *buffer, uint8_t pathLen, uint16_t maxItems, char *extFilter)
{
	DIR currentDir;
	FILINFO info;
	uint16_t item=0;
	
	//start with root directory
	debug_msg("opening dir "); debug_txt((char*)dirName); debug_txt(", result: ");
	FRESULT res = f_opendir(&currentDir, dirName);
	fsDispFRESULT(res);
	if(res != FR_OK) return(0);
	
	debug_msg("dir opened");
	
	while ((f_readdir(&currentDir, &info) == FR_OK) && info.fname[0])
	{
		//debug_msg("analyzing item: "); debug_txt(info.fname); debug_txt(" fattrib="); debug8_t(info.fattrib);
		
		
		switch(extFilter[0])
		{
			case '*':	//'*' = all files
			{
				//skip directories
				if(info.fattrib & AM_DIR)
				{
					debug_msg("skipping dir: "); debug_txt(info.fname);
					continue;
				}
				break;
			}
			case (int)NULL:	//NULL = directories
			{
				//skip not-directories
				if((info.fattrib & AM_DIR)==0) continue;
				break;
			}
			default:	//files with selected extensions
			{
				//convert file name to lower case
				strlwr(info.fname);
				
				//convert ext mask to lower case
				strlwr(extFilter);
				
				//skip directories
				if(info.fattrib & AM_DIR)
				{
					//debug_msg("skipping dir: "); debug_txt(info.fname);
					continue;
				}
				
				//filter extensions
				if(strcmp(extFilter,&info.fname[strlen(info.fname)-3]) != 0)
				{
					continue;
				}
				break;
			}
		}
		
		//debug_msg("adding item: "); debug_txt(info.fname);
		
		//copy obtained name to a buffer
		strcpy(buffer,info.fname);
		
		//increment item counter
		item++;
		
		//advance pointer in buffer
		buffer += pathLen;
		
		//check if limit is reached
		if(item >= maxItems) return(item);
		//compare extension
		
		
	}
	
	return(item);
}


void fsListFiles(char *dirPath)
{
	FILINFO info;
	DIR currentDir;
	
	if(f_opendir(&currentDir, dirPath) == FR_OK)
	{
		while ((f_readdir(&currentDir, &info) == FR_OK) && info.fname[0])
		{
			debug_msg(info.fname);
		}
	}
}


FRESULT fsRemove(char* path)
{
	FRESULT res = f_unlink((const char*)path);
	if(res!=FR_OK)
	{
		debug_msg("ERR fsRemove error");
		fsDispFRESULT(res);
	}
	return(res);
}



FRESULT fsMkDir(char* dirPath)
{
	FRESULT res = f_mkdir((const char*)dirPath);
	if(res!=FR_OK)
	{
		debug_msg("ERR fsMkDir error");
		fsDispFRESULT(res);
	}
	return(res);
}




FRESULT fsGetFileSize(char* fileName, uint32_t *size)
{
	FRESULT res = f_stat((const char*)fileName,&fileInfo);
	if(res!=FR_OK)
	{
		*size = 0xFFFFFFFF;
		fsDispFRESULT(res);
		return(res);
	}
	*size = fileInfo.fsize;
	return(res);
}


FRESULT fsRewrite(char* fileName, void *buf, uint16_t len)
{
	//open file
	FRESULT res = f_open(&fileW,(const char*)fileName,FA_WRITE|FA_CREATE_ALWAYS);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRewrite: f_open error");
		fsDispFRESULT(res);
		f_close(&fileW);
		return(res);
	}
	//set cursor in file
	res = f_lseek(&fileW,0);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRewrite: f_lseek error");
		fsDispFRESULT(res);
		f_close(&fileW);
		return(res);
	}
	//write buffer
	UINT written=0;
	res = f_write(&fileW,buf,len,&written);
	if((res != FR_OK)||(written!=len))
	{
		debug_msg("ERR fsRewrite: f_write error");
		fsDispFRESULT(res);
		debug_value("expected len =",len);
		debug_value("bytes actually written =",written);
		f_close(&fileW);
		return(res);
	}
	//close file
	res = f_close(&fileW);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRewrite: f_close error");
		fsDispFRESULT(res);
		return(res);
	}
	
	return(res);
}


//returns number of bytes read
//fileName: a pointer to file name string
//buf: pointer to buffer for data from file
//pos: starting byte in file
//len: number of bytes to read from file
uint16_t fsRead(char* fileName, void *buf, uint32_t pos, uint16_t len)
{
	//open file
	FRESULT res = f_open(&fileR,(const char*)fileName,FA_READ);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRead: f_open error");
		fsDispFRESULT(res);
		f_close(&fileR);
		return(0);
	}
	//set cursor in file
	res = f_lseek(&fileR,pos);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRead: f_lseek error");
		fsDispFRESULT(res);
		f_close(&fileR);
		return(0);
	}
	//read buffer
	UINT bytesRead=0;
	res = f_read(&fileR,buf,len,&bytesRead);
	
	if(res != FR_OK)
	{
		debug_msg("ERR fsRead: f_read error");
		fsDispFRESULT(res);
		f_close(&fileR);
		return(0);
	}
	
	if(bytesRead!=len)
	{
		debug_msg("WRN fsRead: f_read warning");
		fsDispFRESULT(res);
		debug_value("expected len =",len);
		debug_value("bytes actually read =",bytesRead);
		f_close(&fileR);
		return(bytesRead);
	}
	
	//close file
	res = f_close(&fileR);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRead: f_close error");
		fsDispFRESULT(res);
	}
	
	return(bytesRead);
}


inline FRESULT fsWriteLine(char* fileName, char *buf)
{
	FRESULT res;
	res = fsAppend((char*)fileName,buf,strlen((char*)buf));
	if(res != FR_OK) return(res);
	res = fsAppend(fileName,(uint8_t*)"\r\n",2);
	return(res);
}


inline FRESULT fsAppend(char* fileName, void *buf, uint16_t len)
{
	FRESULT res = fsReplace(fileName,buf,F_LSEEK_EOF,len);
	if(res != FR_OK)
	{
		debug_msg("ERR fsAppend: fsReplaceBufInFile error");
		fsDispFRESULT(res);
		return(res);
	}
	return(res);
}


FRESULT fsReplace(char* fileName, void *buf, uint32_t pos, uint16_t len)
{
	//open file
	FRESULT res = f_open(&fileW,(const char*)fileName,FA_WRITE|FA_OPEN_ALWAYS);
	if(res != FR_OK)
	{
		debug_msg("ERR fsReplaceBufInFile: f_open error");
		fsDispFRESULT(res);
		return(res);
	}
	//set cursor in file
	res = f_lseek(&fileW,pos);
	if(res != FR_OK)
	{
		debug_msg("ERR fsReplaceBufInFile: f_lseek error");
		fsDispFRESULT(res);
		f_close(&fileW);
		return(res);
	}
	//write buffer
	UINT written=0;
	res = f_write(&fileW,buf,len,&written);
	if((res != FR_OK)||(written!=len))
	{
		debug_msg("ERR fsReplaceBufInFile: f_write error");
		fsDispFRESULT(res);
		debug_value("expected len =",len);
		debug_value("bytes actually written =",written);
		f_close(&fileW);
		return(res);
	}
	//close file
	res = f_close(&fileW);
	if(res != FR_OK)
	{
		debug_msg("ERR fsReplaceBufInFile: f_close error");
		fsDispFRESULT(res);
		return(res);
	}
	
	return(res);
}





//=================================================================================================
//								DEBUG UTILS
//=================================================================================================


void fsDispFRESULT(FRESULT res)
{
	debug_txt("FRESULT=");
	switch(res)
	{
		case FR_OK: debug_txt("FR_OK"); break;
		case FR_NOT_READY: debug_txt("FR_NOT_READY"); break;
		case FR_NO_FILE: debug_txt("FR_NO_FILE"); break;
		case FR_NO_PATH: debug_txt("FR_NO_PATH"); break;
		case FR_INVALID_NAME: debug_txt("FR_INVALID_NAME"); break;
		case FR_DENIED: debug_txt("FR_DENIED"); break;
		case FR_RW_ERROR: debug_txt("FR_RW_ERROR"); break;
		case FR_WRITE_PROTECTED: debug_txt("FR_WRITE_PROTECTED"); break;
		case FR_NOT_ENABLED: debug_txt("FR_NOT_ENABLED"); break;
		case FR_NO_FILESYSTEM: debug_txt("FR_NO_FILESYSTEM"); break;
		default: debug_txt("!!! UNDEFINED CODE");
	}
}


//=================================================================================================
//								INITIALIZATION UTILS
//=================================================================================================

FRESULT fsInit(void)
{
	/* Enable FatFs module */
	memset(&fs, 0, sizeof(FATFS));
	
	debug_msg("ff f_mountdrv...");
	FRESULT mountRetval=f_mount(0,&fs);
	fsDispFRESULT(mountRetval);
	return(mountRetval);
}


