#include "common.h"
#include "sd.h"
#include "mmu/mmu.h"


static uint32_t* translationTable = (uint32_t*)0x201FC000;


void initTranslationTable(uint32_t *pTranslationTable)
{
	uint32_t i;
	
	// Reset table entries
	for (i = 0; i < 4096; ++i)
		pTranslationTable[i] = 0;
	
	//vectors (internal SRAM0 after remap)
	pTranslationTable[0x000] =
		(0x000 << 20) |		// Physical Address
		(1 << 10) |			// Access is free
		(1 << 11) |			// 
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section
	
	//default location of internal SRAM0
	pTranslationTable[0x002] =
		(0x002 << 20) |		// Physical Address
		(1 << 10) |			// Access is free
		(1 << 11) |			// 
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section
	
	//default location of internal SRAM1
	pTranslationTable[0x003] =
		(0x003 << 20) |		// Physical Address
		(1 << 10) |			// Access is free
		(1 << 11) |			// 
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section
	
	//the very first 1MB of SDRAM at PA=0x20000000
	pTranslationTable[0x200] =
		(0x200 << 20) |		// Physical Address
		(1 << 10) |			// Access is free
		(1 << 11) |			// 
		(15 << 5) |			// Domain 15
		(1 << 2) |			// Bufferable
		(1 << 3) |			// Cachable
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section
	
	//noncachable area of SDRAM
	for(i=0x201;i<0x210;i++)
	{
		pTranslationTable[i] =
			(i << 20) |			// Physical Address
			(1 << 10) |			// Access is free
			(1 << 11) |			// 
			(15 << 5) |			// Domain 15
			(1 << 4) |
			0x2;				// Set as 1 Mbyte section
	}
	
	//peripherals
	pTranslationTable[0xFFF] =
		(0xFFF << 20) |		// Physical Address
		(1 << 10) |			// Access is free
		(1 << 11) |			// 
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section*/
	
}


void mmuInit(void)
{
	//initialize the Translation Table
	initTranslationTable(translationTable);
	
	//Set the base address of Translation Table
	AT91F_ARM_WriteTTB((uint32_t)translationTable);
	
	//Program the domain access register
	//domain 15: accesses are checked against AP settings (client)
	AT91F_ARM_WriteDomain(0x40000000);
	
	//Enable the MMU
	debug_msg("Enabling the MMU...");
	AT91F_EnableMMU();
	debug_txt("ok");	//display message if it works or hangs...
	
	debug_msg("Enabling caches...");
	AT91F_EnableICache();
	AT91F_EnableDCache();
	debug_txt("ok");
	
	
}



void init(void)
{
	// enable reset-key on board 
	AT91C_BASE_RSTC->RSTC_RMR = (0xA5000000 | AT91C_RSTC_URSTEN);
	
	// enable peripheral clock for PIO controller
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA) | (1<<AT91C_ID_PIOB) | (1<<AT91C_ID_PIOC);
	
	LEDS_INIT;
	debug_init_default();
	
	debug_msg("AT91SAM9260 MMU tester");
	
	mmuInit();
}


void doSomeDummyProcessing(void)
{
	volatile double someDouble1 = 0;
	volatile double someDouble2 = 2000;
	volatile double someDouble3 = 3000;
	volatile uint32_t i;
	for(i=0;i<0x4FFFF;i++)
	{
		someDouble1 = someDouble2 + someDouble3;
		someDouble1 = someDouble1 * 12345;
		someDouble2 = someDouble3 - i;
	}
}


void timerStart(void)
{
	//reset RTT count and set prescaler
	//one unit is 1/1024 sec
	AT91C_BASE_RTTC->RTTC_RTMR = AT91C_RTTC_RTTRST | (32768/1024);
}


uint32_t timerStop(void)
{
	//just read RTT
	return(AT91C_BASE_RTTC->RTTC_RTVR);
}


int main( void )
{
	init();
	
	debug_msg("entering main loop");
	
	for(;;)
	{
		LED_B_TOGGLE;
		timerStart();
		doSomeDummyProcessing();
		uint32_t time = timerStop();
		debug_msg("time taken: "); debug32_t(time);
	}
	return(0);
}

/* test results:
	no cache:		000011AB = 4523
	only DCache:	00000C8C = 3212
	only ICache:	000004A9 = 1193
	both DCaches:	0000023A = 570
	
	time without cache / time with both caches on = 4523 / 570 = ~8 times faster with cache!
	...not to mention free data bus!
*/

