#include "common.h"
#include "dataflash.h"
#include "xmodem.h"
#include "mmu/mmu.h"

//Bootloader downloads program code to SDRAM:
//- from DataFlash AT45DB chip if BOOT_FROM_DATAFLASH macro is true or
//- from serial port (Xmodem protocol) if BOOT_FROM_DATAFLASH macro is false.
//After downloading, the program is started from SDRAM


//BOOT_FROM_DATAFLASH macro may store constant or non-constant value.
//It is used to check user request only once in main -
//set this macro to anything you want (which may be true or false)
//for example checking if a button is pressed
#define BOOT_FROM_DATAFLASH		((AT91C_BASE_PIOC->PIO_PDSR & AT91C_PIO_PC15)==0)

//use this macro to switch on or off initialization of cache memories
//if this macro is true, both caches are initialized
#define PREPARE_CACHES			0

#define BYTES_TO_COPY_FROM_DATAFLASH	(100*1024)

#define READ(peripheral, register)          (peripheral->register)
#define WRITE(peripheral, register, value)  (peripheral->register = value)

void init(void)
{
	// enable peripheral clock for PIO controller
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA) | (1<<AT91C_ID_PIOB) | (1<<AT91C_ID_PIOC);
	debug_init_default();
	debug_msg("SDRAM Bootloader");
}


void sdramInit(void)
{
	volatile uint32_t i;
	volatile uint32_t *pSdram = (uint32_t *) AT91C_EBI_SDRAM;
	
	// Enable EBI chip select for the SDRAM
	*AT91C_MATRIX_EBI |= AT91C_MATRIX_CS1A_SDRAMC;
	
	AT91C_BASE_SDRAMC->SDRAMC_CR = 		AT91C_SDRAMC_NC_9
                        				| AT91C_SDRAMC_NR_13 
                        				| AT91C_SDRAMC_CAS_2 
                        				| AT91C_SDRAMC_NB_4_BANKS
                        				| AT91C_SDRAMC_DBW_16_BITS	//16-bit bus (1xSDRAM)
                        				| AT91C_SDRAMC_TWR_2
                        				| AT91C_SDRAMC_TRC_7
                        				| AT91C_SDRAMC_TRP_2
                        				| AT91C_SDRAMC_TRCD_2
                        				| AT91C_SDRAMC_TRAS_5
                        				| AT91C_SDRAMC_TXSR_8;
	
	for (i = 0; i < 1000; i++);
	
	AT91C_BASE_SDRAMC -> SDRAMC_MR = AT91C_SDRAMC_MODE_NOP_CMD;	// Perform NOP
	pSdram[0] = 0x00000000;
	
	WRITE(AT91C_BASE_SDRAMC, SDRAMC_MR, AT91C_SDRAMC_MODE_PRCGALL_CMD);	// Set PRCHG AL
	pSdram[0] = 0x00000000;						// Perform PRCHG
	
	for (i = 0; i < 10000; i++);
	
	AT91C_BASE_SDRAMC->SDRAMC_MR = AT91C_SDRAMC_MODE_RFSH_CMD;
	pSdram[1] = 0x00000001;
	AT91C_BASE_SDRAMC->SDRAMC_MR = AT91C_SDRAMC_MODE_RFSH_CMD;
	pSdram[2] = 0x00000002;
	
	AT91C_BASE_SDRAMC->SDRAMC_MR = AT91C_SDRAMC_MODE_RFSH_CMD;
	pSdram[3] = 0x00000003;
	
	AT91C_BASE_SDRAMC->SDRAMC_MR = AT91C_SDRAMC_MODE_RFSH_CMD;
	pSdram[4] = 0x00000004;
	
	AT91C_BASE_SDRAMC->SDRAMC_MR = AT91C_SDRAMC_MODE_RFSH_CMD;
	pSdram[5] = 0x00000005;
	
	AT91C_BASE_SDRAMC->SDRAMC_MR = AT91C_SDRAMC_MODE_RFSH_CMD;
	pSdram[6] = 0x00000006;
	
	AT91C_BASE_SDRAMC->SDRAMC_MR = AT91C_SDRAMC_MODE_RFSH_CMD;
	pSdram[7] = 0x00000007;
	
	AT91C_BASE_SDRAMC->SDRAMC_MR = AT91C_SDRAMC_MODE_RFSH_CMD;
	pSdram[8] = 0x00000008;
	
	AT91C_BASE_SDRAMC->SDRAMC_MR = AT91C_SDRAMC_MODE_LMR_CMD;
	pSdram[9] = 0xAAAAAAAA;
	
	AT91C_BASE_SDRAMC->SDRAMC_TR = ((MCK * 7) / 1000000);		// Set Refresh Timer
	
	AT91C_BASE_SDRAMC->SDRAMC_MR = AT91C_SDRAMC_MODE_NORMAL_CMD;	// Set Normal mode
	pSdram[10] = 0x00000000;						// Perform Normal mode
}


static TXmodemBlock block;



void (*addressToStart)();

int main( void )
{
	const uint8_t APPLICATION_FIRST_BLOCK = 0x0A;
	const uint16_t DF_BUFFER_SIZE = 528;
	volatile uint8_t *pSdram = (uint8_t*) AT91C_EBI_SDRAM;
	
	init();
	
	dataflashInit();
	
	sdramInit();
	
	#if PREPARE_CACHES
	AT91F_InitMMU();
	#endif
	
	uint8_t bootingFromDataflash;
	
	if(BOOT_FROM_DATAFLASH)
	{
		debug_msg("Booting from DataFlash. Init...");
		bootingFromDataflash = 1;
		
		//wait for good dataflash status
		uint8_t dfStatus;
		do
		{
			dfStatus = dataflashReadStatusRegister();
		}while((dfStatus&0xC0) != 0x80);
		
		debug_msg("downloading...");
		
	}
	else
	{
		debug_msg("SDRAM Bootloader. Send file to SDRAM\r\n");
		bootingFromDataflash = 0;
	}
	
	pSdram = (uint8_t*) AT91C_EBI_SDRAM;
	
	if(bootingFromDataflash)
	{
		//initialize read from dataflash
		dataflashStartNewStream();
		
		for(uint16_t i=0;i<(DF_BUFFER_SIZE*APPLICATION_FIRST_BLOCK);i++)
		{
			//dummy reads
			dataflashReadStream();
		}
		
		//copy 200KB of application to SDRAM
		for(uint32_t i=0; i<BYTES_TO_COPY_FROM_DATAFLASH; i++)
		{
			pSdram[i] = dataflashReadStream();
			//debug_chr('*');
		}
		
	} //if(bootingFromDataflash)
	else 
	{
		pSdram = (uint8_t*) AT91C_EBI_SDRAM;
		for(;;)
		{
			uint8_t xmodemStatus = xmodemDownloadBlock((TXmodemBlock*)&block);
			
			//try again if block was bad
			if(xmodemStatus == 0) continue;
			
			for(uint16_t i=0;i<XMODEM_PAYLOAD_LENGTH;i++)
			{
				*pSdram = block.payload[i];
				pSdram++;
			} //for(i=0;i<XMODEM_PAYLOAD_LENGTH;i+=2)
			if(xmodemStatus == 2) break;
		} //for(;;)
	} //else [if(bootingFromDataflash)]
	
	debug_msg("sdram contents:");
	
	uint8_t *codePtr = (uint8_t*) AT91C_EBI_SDRAM;
	uint32_t address = 0;
	debug_msg("from address "); debug32_t(address);
	for(uint16_t i=0;i<0x80;i++)
	{
		if((i%16)==0) debug_msg("");
		debug8_t(codePtr[address++]); debug_chr(' ');
	}
	
	debug_msg("ready!");
	
	dataflashDeactivate();
	
	addressToStart = (void(*))AT91C_EBI_SDRAM;
	
	//a very unprofessional delay...
	for(uint32_t i=0;i<0x20000;i++) asm volatile("nop"::);
	
	#if PREPARE_CACHES
		//enable caches after programming the memory
		//comment/uncomment one or both of them to try the performance!
		AT91F_EnableICache();	//enables Instruction Cache
		AT91F_EnableDCache();	//enables Data Cache
	#endif
	
	//jump to the program
	addressToStart();
	
	return(0);
}


