#include "common.h"
#include "sd.h"
#include "fat/fs_tools.h"
#include "mmu/mmu.h"

static FIL file;

static uint32_t* translationTable = (uint32_t*)0x201FC000;

/*
for domain and access permission settings see page 88 of ARM926EJ-S TRM
*/

void initTranslationTable(uint32_t *pTranslationTable)
{
	uint32_t i;
	
	// Reset table entries
	for (i = 0; i < 4096; ++i)
		pTranslationTable[i] = 0;
	
	//vectors (internal SRAM0 after remap)
	pTranslationTable[0x000] =
		(0x000 << 20) |		// Physical Address
		(1 << 10) |			// Access in SVC: R/W, in USR: none
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section
	
	//default location of internal SRAM0
	pTranslationTable[0x002] =
		(0x002 << 20) |		// Physical Address
		(1 << 10) |			// Access in SVC: R/W, in USR: none
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section
	
	//default location of internal SRAM1
	pTranslationTable[0x003] =
		(0x003 << 20) |		// Physical Address
		(1 << 10) |			// Access in SVC: R/W, in USR: none
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section
	
	//the very first 1MB of SDRAM at PA=0x20000000
	pTranslationTable[0x200] =
		(0x200 << 20) |		// Physical Address
		(1 << 10) |			// Access in SVC: R/W, in USR: none
		(15 << 5) |			// Domain 15
		(1 << 2) |			// Bufferable
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section
	
	pTranslationTable[0x201] =
		(0x201 << 20) |		// Physical Address
		(1 << 10) |			// Access in supervisor mode
		//(1 << 11) |
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section
	
	//User area: PA from 0x40000000
	pTranslationTable[0x400] =
		(0x201 << 20) |		// Physical Address
		(1 << 10) |			// Access is free
		(1 << 11) |			// 
		(15 << 5) |			// Domain 15
		(1 << 4) |
		(1 << 3) |			// Cachable
		(1 << 2) |			// Bufferable
		0x2;				// Set as 1 Mbyte section
	
	//noncachable area of SDRAM
	for(i=0x202;i<0x210;i++)
	{
		pTranslationTable[i] =
			(i << 20) |			// Physical Address
			(1 << 10) |			// Access in SVC: R/W, in USR: none
			(15 << 5) |			// Domain 15
			(1 << 4) |
			(1 << 2) |			// Bufferable
			0x2;				// Set as 1 Mbyte section
	}
	
	//peripherals
	pTranslationTable[0xFFF] =
		(0xFFF << 20) |		// Physical Address
		(1 << 10) |			// Access in SVC: R/W, in USR: none
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section*/
	
}


void mmuInit(uint32_t *pTranslationTable)
{
	//initialize the Translation Table
	initTranslationTable(pTranslationTable);
	
	//Set the base address of Translation Table
	AT91F_ARM_WriteTTB((uint32_t)pTranslationTable);
	
	//Program the domain access register
	//domain 15: accesses are checked (client)
	AT91F_ARM_WriteDomain(0x40000000);
	
	//Enable the MMU
	debug_msg("Enabling the MMU...");
	AT91F_EnableMMU();
	debug_txt("ok");	//display message if it works or hangs...
	
	debug_msg("Enabling caches...");
	AT91F_EnableICache();
	AT91F_EnableDCache();
	debug_txt("ok");
	
	
}


//reads and returns CPSR register
inline uint32_t getCpsr(void)
{
	uint32_t res;
	asm volatile("mrs %0,cpsr":"=r"(res));
	return(res);
}


inline uint32_t getPc(void)
{
	register volatile uint32_t pc asm("r15");
	return(pc);
}




void swiHandler(void)
{
	
	//address of SWI instruction is stored in R12
	register volatile uint32_t *swiInstruction asm ("r12");
	uint32_t swiNumber = *swiInstruction & 0x00FFFFFF;
	debug_msg("SWI Exception! SWI number="); debug32_t(swiNumber);
	
	switch(swiNumber)
	{
		case 0x1001: LED_R_TOGGLE; break;
		case 0x1002: LED_G_TOGGLE; break;
		case 0x1003: LED_Y_TOGGLE; break;
		case 0x1004: LED_B_TOGGLE; break;
	}
}


//pass address of bad instruction as an argument
void displayAbortInfo(uint32_t *badInstruction)
{
	debug_msg("Instruction that caused abort was: ");
		debug32_t((uint32_t)*badInstruction);
	debug_txt(", at address: ");
		debug32_t((uint32_t)badInstruction);
	debug_msg("Data Fault Status: "); debug32_t(AT91F_ARM_ReadDataFSR());
	debug_msg("Prefetch Fault Status: "); debug32_t(AT91F_ARM_ReadPrefetchFSR());
	debug_msg("Fault Address Register: "); debug32_t(AT91F_ARM_ReadFAR());
	debug_msg("press any key to continue...");
	debug_waitkey();
}

//handles prefetch abort generated by user program
void userProgramprefetchAbortHandler(void)
{
	register volatile uint32_t *badInstruction asm ("r12");
	uint32_t *temp = (uint32_t*)badInstruction;
	debug_msg("*** User program caused Prefetch Abort!");
	displayAbortInfo(temp);
}

//handles data abort generated by user program
void userProgramDataAbortHandler(void)
{
	register volatile uint32_t *badInstruction asm ("r12");
	uint32_t *temp = (uint32_t*)badInstruction;
	debug_msg("*** User program caused Data Abort!");
	displayAbortInfo(temp);
}

//handles data abort generated by supervisor program
void supervisorDataAbortHandler(void)
{
	register volatile uint32_t *badInstruction asm ("r12");
	uint32_t *temp = (uint32_t*)badInstruction;
	debug_msg("*** Supervisor program caused Data Abort!");
	displayAbortInfo(temp);
}

//handles prefetch abort generated by supervisor program
void supervisorPrefetchAbortHandler(void)
{
	register volatile uint32_t *badInstruction asm ("r12");
	uint32_t *temp = (uint32_t*)badInstruction;
	debug_msg("*** Supervisor program caused Prefetch Abort!");
	displayAbortInfo(temp);
}



void init(void)
{
	// enable reset-key on board 
	AT91C_BASE_RSTC->RSTC_RMR = (0xA5000000 | AT91C_RSTC_URSTEN);
	
	// enable peripheral clock for PIO controller
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA) | (1<<AT91C_ID_PIOB) | (1<<AT91C_ID_PIOC);
	
	LEDS_INIT;
	debug_init_default();
	
	debug_msg("AT91SAM9260 MMU tester");
	
	mmuInit(translationTable);
	
	fsInit();
	
}



void runProgramFromSD(const char *fileName)
{
//	debug_msg("running program: "); debug_txt((char*)fileName);

	FRESULT res;
	res = f_open(&file,fileName,FA_READ);
	if(res != FR_OK)
	{
		debug_msg("could not open the file, result="); debug8_t(res);
		return;
	}
	
	debug_msg("disabling ICache...");
	AT91F_DisableICache();
	debug_txt("OK");
	
	debug_msg("resetting ICache...");
	AT91F_ARM_InvalidateICache();
	debug_txt("OK");
	
	debug_msg("disabling DCache...");
	AT91F_DisableDCache();
	debug_txt("OK");
	
	debug_msg("resetting DCache...");
	AT91F_CleanDCache();
	AT91F_ARM_DrainWriteBuffer();
	AT91F_ARM_InvalidateDCache();
	debug_txt("OK");
	
	uint32_t *programArea = (uint32_t*)0x40000000;
	
	UINT bytesRead=0;
	res = f_read(&file,(void*)programArea,0xFFFF,&bytesRead);
	if(res != FR_OK)
	{
		debug_msg("program copying error, result="); debug8_t(res);
		return;
	}
	debug_value("bytes read: ", bytesRead);
	
	f_close(&file);
	
	//debug_msg("CPSR="); debug32_t(getCpsr());
	
	debug_msg("reenabling Cache...");
	AT91F_EnableICache();
	AT91F_EnableDCache();
	debug_txt("OK");
	
	debug_msg("starting user program");
	
	asm volatile(
	"stmfd sp!, {r0,r1} \n\t"		//save r0 and r1 on supervisor's stack
	"ldr r0,=0x40000000 \n\t"		//set address of user program to jump in r0
	"mov r1,pc \n\t"				//compute go-back address and store it in r1
	"add r1,r1,#12 \n\t"			//address to go back is current PC+12 bytes
	"stmfd sp!, {r1-r12,lr} \n\t"	//save registers to supervisor's stack
	"msr cpsr,0x10 \n\t"			//switch to user mode
	"mov pc, r0 \n\t"				//jump to address of user program
	
	//this is the first instruction after go back
	"msr CPSR_c, 0x13 \n\t"			//clear I and F flags (SWI exception sets I)
	"ldmia sp!, {r0,r1} \n\t"		//restore other registers
	);
	
	debug_msg("back in main program");
}



int main( void )
{
	init();
	
	debug_msg("entering main loop");
	
	for(;;)
	{
		uint8_t key = debug_inkey();
		
		switch(key)
		{
			case '1':
				debug_msg("generating SWI 1");
				asm volatile("SWI"::);
				break;
			case '2':
				debug_msg("generating SWI 2");
				asm volatile("SWI 2"::);
				break;
			case '3':
				debug_msg("generating SWI 0xABCD");
				asm volatile("SWI 0xABCD"::);
				break;
			case 'R':
				debug_msg("running a program, enter file name: ");
				char programName[8+1+3+1];
				debug_input_string(programName,8+1+3+1);
				runProgramFromSD(programName);
				break;
		}
		
	}
	return(0);
}

