#include "common.h"
#include "sd.h"

#define MCI_PIO			AT91C_BASE_PIOA		//PIO base address (where MCI pins are muxed)
#define MCI_PSR			PIO_BSR				//peripheral select register A or B
#define MCI_MCCK_MASK	AT91C_PA8_MCCK
#define MCI_MCDB0_MASK	AT91C_PA0_MCDB0
#define MCI_PINS_MASK	( \
	MCI_MCDB0_MASK | AT91C_PA1_MCCDB | AT91C_PA3_MCDB3 |\
	AT91C_PA4_MCDB2 | AT91C_PA5_MCDB1 | MCI_MCCK_MASK )	//pins used by MCI controller
#define MCI				AT91C_BASE_MCI		//Base address of MCI to use
#define MCI_ID			AT91C_ID_MCI		//ID of MCI to use

//define which leds blink during write and read
//if you don't want any leds, just leave these 4 defs empty
#define LED_WR_ON			//LED_R_ON
#define LED_WR_OFF			//LED_R_OFF
#define LED_RD_ON			//LED_Y_ON
#define LED_RD_OFF			//LED_Y_OFF

//if set to ZERO, variable (card dependent) block size will be used
//if set to NONZERO, fixed (card independent) block size will be used
#define FIXED_BLOCK_SIZE	512



//define error masks
#define MCI_ERRORS_MASK	(AT91C_MCI_RINDE | AT91C_MCI_RDIRE   \
		| AT91C_MCI_RENDE | AT91C_MCI_RTOE | AT91C_MCI_DCRCE \
		| AT91C_MCI_DTOE | AT91C_MCI_OVRE | AT91C_MCI_UNRE)


//don't care argument
#define ARG_DONT_CARE		0x00000000

#define SD_SEND_RELATIVE_ADDR			3
#define SD_SEND_IF_COND					8
#define SD_SWITCH						6
#define SD_APP_SET_BUS_WIDTH			6
#define SD_APP_SEND_NUM_WR_BLKS			22
#define SD_APP_OP_COND					41
#define SD_APP_SEND_SCR					51
#define MMC_APP_CMD						55
#define	MMC_GO_IDLE_STATE				0
#define MMC_ALL_SEND_CID				2
#define MMC_SET_RELATIVE_ADDR			3
#define MMC_SELECT_CARD					7
#define MMC_SEND_CSD					9
#define MMC_SEND_CID					10
#define MMC_STOP_TRANSMISSION			12
#define MMC_SEND_STATUS					13
#define MMC_WRITE_BLOCK					24
#define MMC_SET_BLOCKLEN				16
#define MMC_READ_SINGLE_BLOCK			17
#define MMC_READ_MULTIPLE_BLOCK			18
#define MMC_WRITE_MULTIPLE_BLOCK		25
#define MMC_ERASE_WR_BLK_START			32
#define MMC_ERASE_WR_BLK_END			33
#define MMC_ERASE						38
#define SD_APP_SET_WR_BLK_ERASE_COUNT	23


//Relative Card Address (RCA)
static uint32_t rca;

//block size in bytes
static uint32_t blockSize;

//number of blocks on card
static uint32_t numberOfBlocks;

static uint8_t sdhcFlag;

//private functions
static uint8_t sdHWinit(void);
static uint8_t sdCardInit(void);
static uint8_t sdCommand(uint32_t command, uint32_t argument);
static uint8_t sdResponse48(uint32_t *buf);
static uint8_t sdResponse136(uint32_t *buf);
static uint8_t sdBusyWait(void);
static uint32_t sdReadStatus(void);

//single, no-dma block operations
static uint8_t sdReadBlock(void *buffer, uint32_t lba);
static uint8_t sdWriteBlock(void *buffer, uint32_t lba);


void sdDispResponse(uint32_t *pResp)
{
	debug_msg("response: ");
//	uint8_t *pResp8 = (uint8_t*)pResp;
	for(uint8_t i=0;i<5;i++)
	{
		debug32_t(*pResp++);
		debug_chr(' ');
	}
}

static uint8_t sdHWinit(void)
{
	//enable MCI's clock
	AT91C_BASE_PMC->PMC_PCER = (1<<MCI_ID);
	
	//configure PIOs to be controlled by MCI
	MCI_PIO->PIO_PDR = MCI_PINS_MASK;
	
	//select peripheral for MCI pins
	MCI_PIO->MCI_PSR = MCI_PINS_MASK;
	
	//enable internal pullupz
	MCI_PIO->PIO_PPUER = MCI_PINS_MASK;
	
	//configure MCI regs
	
	//perform software reset
	MCI->MCI_CR = AT91C_MCI_SWRST;
	
	//enable MCI, disable power saving mode
	MCI->MCI_CR = AT91C_MCI_MCIEN;
	
	//prepare masks
	const uint8_t CLKDIV = 0xFF;
	const uint32_t MCI_CLKDIV_MASK = (uint32_t)CLKDIV;
	const uint16_t BLKLEN = 512;	//only an initial, default value
	const uint32_t MCI_BLKLEN_MASK = (uint32_t)(BLKLEN << 16);
	
	//set MCI Mode Register
	MCI->MCI_MR = MCI_BLKLEN_MASK | AT91C_MCI_PDCPADV | MCI_CLKDIV_MASK | AT91C_MCI_RDPROOF | AT91C_MCI_WRPROOF;
	
	const uint8_t DTOCYC = 0x0F;
	
	//set timeout stuff
	MCI->MCI_DTOR = DTOCYC | AT91C_MCI_DTOMUL_1048576;
	
	return(OK);
}


static uint8_t sdCardInit(void)
{
	LED_WR_ON;
	LED_RD_ON;
	
	//universal buffer for responses
	uint32_t response[5];
	
	//select slot B...
	const uint8_t SDCSEL_SLOT_B = 1;
	//...in MCI SDCR register
	MCI->MCI_SDCR = SDCSEL_SLOT_B | AT91C_MCI_SCDBUS;		//AT91C_MCI_SCDBUS means 4-bit bus
	
	//initialization command (74 clock cycles for initialization sequence)
	sdCommand(AT91C_MCI_SPCMD_INIT, ARG_DONT_CARE);
	
	//issue CMD0 (something like card reset)
	sdCommand(MMC_GO_IDLE_STATE,ARG_DONT_CARE);
	
	//Issue CMD8 with 10101010 arg and VHS=1
	if(sdCommand(SD_SEND_IF_COND | AT91C_MCI_MAXLAT | AT91C_MCI_RSPTYP_48,0x000001AA)==OK)
	{
		debug_msg("sdInit: New card in slot (response to CMD8 was OK)");
	}
	else
	{
		debug_msg("sdInit: Old card in slot (no response to CMD8)");
	}
	
	//wait for initialization and enter "ready state"
	if(sdBusyWait()!=OK)
	{
		debug_msg("sdBusyWait returned FAIL");
		return(FAIL);
	}

	//request CID and enter "identification state"
	sdCommand(MMC_ALL_SEND_CID | AT91C_MCI_MAXLAT | AT91C_MCI_RSPTYP_136, ARG_DONT_CARE);
	sdResponse136(response);
	
	//issue CMD3 and enter "standby state"
	sdCommand(MMC_SET_RELATIVE_ADDR | AT91C_MCI_MAXLAT | AT91C_MCI_RSPTYP_48, ARG_DONT_CARE);
	sdResponse48(response);
	
	//extract relative card address and copy it to a global variable
	rca = (response[0] >> 16) & 0xFFFF;
	
	//request CSD from Card
	sdCommand(MMC_SEND_CSD | AT91C_MCI_MAXLAT | AT91C_MCI_RSPTYP_136, rca<<16);
	sdResponse136(response);
	
	//compute block length and number of blockz
	//example of returned CSD stuff (SD 256MB)
	//005E0032 1F5983CF EDB6BF87 964000B5 005E0032
	//            ^ ^^^ ^
	//      bl.len| |||C_SIZE
	/*
	  4GB SDHC:
	  400E0032 5B590000 1DE77F80 0A4000D5 400E0032
	  bl.len is fixed to 0x9 and it means 512B/block
	*/
	
	uint8_t *pResp = (uint8_t*)response;
	
	//recognize version of CSD data
	if( (response[0] & 0x40000000) == 0x40000000 )
	{
		//"new" CSD (v2.0) detected
		
		//additionally check if card (probably SDHC) has 512B blocks
		if( (response[1] & 0x00090000) != 0x00090000 )
		{
			debug_msg("sdInit, SDHC card, READ_BL_LEN != 0x9");
			return(FAIL);
		}
		else
		{
			//update global variable
			blockSize = 512;
		}
		
		//compute size and update global "numberOfBlocks"
		uint32_t c_size = ((response[1]<<16) | (response[2]>>16)) & 0x3FFFFF;
		numberOfBlocks = (c_size + 1) * 1024;
		
	}
	else
	{
		//"old" CSD detected
		
		//compute size and update globals
		
		uint32_t size;
		size = pResp[5] & 0x03;
		size <<= 10;
		size |= (uint16_t)pResp[4]<<2;
		size |= pResp[11]>>6;
		
		uint16_t sizeMultiplier = 1 << (2+ ( ( (pResp[10] & 0x03)<<1 ) | (pResp[9] >> 7) ) );
		
		#if FIXED_BLOCK_SIZE == 0
		blockSize = 1 << ((response[1]>>16) & 0x0F);
		#else
		blockSize = FIXED_BLOCK_SIZE;
		#endif
		
		numberOfBlocks=sizeMultiplier*(size+1);
	}
	
	//when CSD reg is read, we may set the full speed (yeah!)
	uint8_t clkdiv = 0x01;
	if(sdhcFlag)
		clkdiv = 0;
	else
		clkdiv = 1;
	uint32_t MCI_CLKDIV_MASK = (uint32_t)clkdiv;
	uint32_t mci_BLOCKLEN_mask = (uint32_t)(blockSize << 16);
	
	//set MCI Mode Register
	MCI->MCI_MR = mci_BLOCKLEN_mask | AT91C_MCI_PDCPADV | MCI_CLKDIV_MASK | AT91C_MCI_RDPROOF | AT91C_MCI_WRPROOF;
	
	//select card (put it into the transfer state)
	sdCommand(MMC_SELECT_CARD | AT91C_MCI_RSPTYP_48 | AT91C_MCI_MAXLAT, rca<<16);
	sdResponse48(response);
	
	//set 4-bit bus width (sequence of 2 commands)
	sdCommand(MMC_APP_CMD | AT91C_MCI_MAXLAT | AT91C_MCI_RSPTYP_48, rca<<16);
	sdResponse48(response);
	sdCommand(SD_APP_SET_BUS_WIDTH | AT91C_MCI_MAXLAT | AT91C_MCI_RSPTYP_48, 2);
	sdResponse48(response);
	
	LED_WR_OFF;
	LED_RD_OFF;
	
	return(OK);
}


static uint32_t sdReadStatus(void)
{
	sdCommand(MMC_SEND_STATUS | AT91C_MCI_MAXLAT | AT91C_MCI_RSPTYP_48, rca<<16);
	uint32_t response[2];
	sdResponse48(response);
	return(response[0]);
}


uint8_t sdReadBlocks(void *buffer, uint32_t lba, uint16_t blocksToRead)
{
	LED_RD_ON;
	uint8_t *pBuf;
	pBuf = (uint8_t*)buffer;
	while(blocksToRead--)
	{
		sdReadBlock(pBuf,lba);
		pBuf+=blockSize;
		lba++;
	}
	LED_RD_OFF;
	
	return(OK);
}

static uint8_t sdReadBlock(void *buffer, uint32_t lba)
{
	//disable PDC mode (if enabled)
	MCI->MCI_MR &= ~AT91C_MCI_PDCMODE;
	
	//set block len in bytes
	MCI->MCI_BLKR = (uint32_t)(blockSize)<<16;
	
	uint32_t byteAddress = lba * blockSize;
	
	uint32_t response[2];
	
	if(sdhcFlag)
	{
		sdCommand(MMC_READ_SINGLE_BLOCK | AT91C_MCI_MAXLAT  | AT91C_MCI_TRDIR | AT91C_MCI_TRCMD_START | AT91C_MCI_RSPTYP_48, lba);
	}
	else
	{
		sdCommand(MMC_READ_SINGLE_BLOCK | AT91C_MCI_MAXLAT  | AT91C_MCI_TRDIR | AT91C_MCI_TRCMD_START | AT91C_MCI_RSPTYP_48, byteAddress);
	}
	sdResponse48(response);	//get response
	
	uint32_t wordsToRead = blockSize / 4;
	uint8_t *pBuf = (uint8_t*)buffer;
	
	while(wordsToRead--)
	{
		while((MCI->MCI_SR & AT91C_MCI_RXRDY)==0);
		uint32_t temp;
		temp = MCI->MCI_RDR;
		for(uint8_t i=0;i<4;i++)
		{
			*pBuf = temp;
			temp = temp>>8;
			pBuf++;
		}
	}
	
	return(OK);
}


uint8_t sdWriteBlocks(void *buffer, uint32_t lba, uint16_t blocksToWrite)
{
	uint8_t *pBuf = (uint8_t*)buffer;
	while(blocksToWrite--)
	{
		if(sdWriteBlock(pBuf,lba)!=OK) {LED_WR_OFF; return(FAIL);}
		pBuf+=blockSize;
		lba++;
	}
	return(OK);
}


static uint8_t sdWriteBlock(void *buffer, uint32_t lba)
{
	LED_WR_ON;
	
	//disable PDC mode (if enabled)
	MCI->MCI_MR &= ~AT91C_MCI_PDCMODE;
	
	//set block len in bytes
	MCI->MCI_BLKR = (uint32_t)(blockSize)<<16 | 1;
	
	uint32_t byteAddress = lba * blockSize;
	uint32_t response[2];
	if(sdhcFlag)
	{
		sdCommand(MMC_WRITE_BLOCK | AT91C_MCI_MAXLAT | AT91C_MCI_TRCMD_START | AT91C_MCI_RSPTYP_48, lba);
	}
	else
	{
		sdCommand(MMC_WRITE_BLOCK | AT91C_MCI_MAXLAT | AT91C_MCI_TRCMD_START | AT91C_MCI_RSPTYP_48, byteAddress);
	}
	sdResponse48(response);
	
	uint32_t wordsToWrite = blockSize / 4;
	uint8_t *pBuf = (uint8_t*)buffer;
	while(wordsToWrite--)
	{
		while(!(MCI->MCI_SR & AT91C_MCI_TXRDY));
		uint32_t temp;
		uint8_t *pTemp = (uint8_t*)&temp;
		for(uint8_t i=0;i<4;i++)
		{
			*pTemp = *pBuf;
			pBuf++;
			pTemp++;
		}
		MCI->MCI_TDR = temp;
	}
	
	while((MCI->MCI_SR & AT91C_MCI_NOTBUSY) == 0);
	
	//read status and poll card's ready bit
	while((sdReadStatus() & (1<<8))==0);
	
	LED_WR_OFF;
	
	return(OK);
}




static uint8_t sdBusyWait(void)
{
	uint32_t response[2];
	for(uint16_t i=0;i<100;i++)
	{
		//after those commands, status will be in response[0]
		
		if(sdCommand(MMC_APP_CMD | AT91C_MCI_MAXLAT | AT91C_MCI_RSPTYP_48,0)!=OK) return(FAIL);
		sdResponse48(response);
		
		if(sdCommand(SD_APP_OP_COND | AT91C_MCI_RSPTYP_48, 0x00FF8000 | (1<<30))!=OK) return(FAIL);
		sdResponse48(response);
		
		delay_ms(10);
		
		if(response[0] & 0x80000000)	//if ready bit is set
		{
			//set SDHC flag
			if((response[0] & 0xC0000000)==0xC0000000) sdhcFlag=1; else sdhcFlag=0;
			return(OK);
		}
	}
	debug_txt("sdhcBusyWait: TIMEOUT!!!");
	return(FAIL);
}


uint8_t sdGetSizeInfo(uint32_t *availableBlocks, uint16_t *sizeOfBlock)
{
	*availableBlocks = numberOfBlocks;
	*sizeOfBlock = blockSize;
	return(OK);
}


uint8_t sdInit(void)
{
	//initialize MCI controller and IO pins
	sdHWinit();
	
	//initialize SD card in slot
	if(sdCardInit()!=OK) return(FAIL);
	
	return(OK);
}

static uint8_t sdResponse48(uint32_t *buf)
{
	for(uint8_t i=0;i<2;i++)
	{
		buf[i] = MCI->MCI_RSPR[i];
	}
	return(OK);
}

static uint8_t sdResponse136(uint32_t *buf)
{
	for(uint8_t i=0;i<5;i++)
	{
		buf[i] = MCI->MCI_RSPR[0];
	}
	return(OK);
}




static uint8_t sdCommand(uint32_t command, uint32_t argument)
{
	MCI->MCI_ARGR = argument;
	MCI->MCI_CMDR = command;
	
	uint32_t status=0, timeTaken = 0;
	
	while((status & AT91C_MCI_CMDRDY)==0)
	{
		status = MCI->MCI_SR;
		timeTaken++;
		if(timeTaken == 0) return(FAIL);
	}
	
	if(status & MCI_ERRORS_MASK)
	{
		//debug_value("sdCommand ERROR, MCI_SR value = ",status);
		return(FAIL);
	}
	
	return(OK);
}




