#include "common.h"
#include "sd.h"
#include "fat/fs_tools.h"
#include "mmu/mmu.h"

static FIL file;

static uint32_t* translationTable = (uint32_t*)0x200FC000;	//aligned to 16KB, each entry represents 1MB block
static uint32_t* finePageTable = (uint32_t*)0x200FB000;	//each entry represents 1KB block

#define TEST_ARRAY_SIZE			200*1024
static volatile uint32_t testArray[TEST_ARRAY_SIZE];

/*
for domain and access permission settings see page 88 of ARM926EJ-S TRM
*/

//initializes a fine page table that represents the 1st 1MB of address space in SDRAM
void initFinePageTable(uint32_t *pFinePageTable)
{
	uint32_t pageBaseAddress = 0x20000000;
	uint32_t entry = 0;
	
	//debug_msg("creating fine page table");
	
	//first 15 large pages
	for(uint32_t i=0;i<15;i++)
	{
		//defines access prermission for each quarter of a 4KB page
		const uint32_t ACCESS_PERMISSION_MASK =
		(0x1 << 4) |		//access for supervisor only
		(0x1 << 6) |		//access for supervisor only
		(0x1 << 8) |		//access for supervisor only
		(0x1 << 10)			//access for supervisor only
		;
		
		//64 entries per page
		for(uint32_t j = 0; j < 64; j++)
		{
			//create fine page descriptor entry
			//debug_msg("index="); debug32_t(entry);
			//debug_txt(", PA="); debug32_t(pageBaseAddress);
			
			pFinePageTable[entry++] =
				pageBaseAddress |
				ACCESS_PERMISSION_MASK |
				//bits 3 and 2 indicate if the area is cachable or bufferable
				1	//this is a large page descriptor
			;
			
		}
		
		pageBaseAddress += 64*1024;
		
		//debug_msg("press any key");
		//debug_waitkey();
	}
	
	//debug_msg("small pages");
	
	//next 11 small pages
	for(uint32_t i=0;i<11;i++)
	{
		//define access prermission for each quarter of a 4KB page
		const uint32_t ACCESS_PERMISSION_MASK =
		(0x1 << 4) |		//access for supervisor only
		(0x1 << 6) |		//access for supervisor only
		(0x1 << 8) |		//access for supervisor only
		(0x1 << 10)			//access for supervisor only
		;
		
		for(uint32_t j = 0; j < 4; j++)
		{
			//debug_msg("index="); debug32_t(entry);
			//debug_txt(", PA="); debug32_t(pageBaseAddress);
			//create fine page descriptor entry
			pFinePageTable[entry++] =
				pageBaseAddress |
				ACCESS_PERMISSION_MASK |
				//bits 3 and 2 indicate if the area is cachable or bufferable
				2	//this is a small page descriptor
			;
		}
		
		pageBaseAddress += 4*1024;
	}

	//debug_msg("press any key, tiny pages");
	//debug_waitkey();
	
	//upper 20 tiny pages
	for(uint32_t i=0;i<20;i++)
	{
		//define access prermission for each quarter of a 4KB page
		const uint32_t ACCESS_PERMISSION_MASK =
		(0x0 << 4)		//no access
		;
		
		//debug_msg("index="); debug32_t(entry);
		//debug_txt(", PA="); debug32_t(pageBaseAddress);
		
		//create fine page descriptor entry
		pFinePageTable[entry++] =
			pageBaseAddress |
			ACCESS_PERMISSION_MASK |
			//bits 3 and 2 indicate if the area is cachable or bufferable
			3	//this is a tiny page descriptor
		;
		
		pageBaseAddress += 1024;
	}

}


void initTranslationTable(uint32_t *pTranslationTable)
{
	uint32_t i;
	
	// Reset table entries
	for (i = 0; i < 4096; ++i)
		pTranslationTable[i] = 0;
	
	//vectors (internal SRAM0 after remap)
	pTranslationTable[0x000] =
		(0x000 << 20) |		// Physical Address
		(1 << 10) |			// Access in SVC: R/W, in USR: none
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section
	
	//default location of internal SRAM0
	pTranslationTable[0x002] =
		(0x002 << 20) |		// Physical Address
		(1 << 10) |			// Access in SVC: R/W, in USR: none
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section
	
	//default location of internal SRAM1
	pTranslationTable[0x003] =
		(0x003 << 20) |		// Physical Address
		(1 << 10) |			// Access in SVC: R/W, in USR: none
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section
	
	//the very first 1MB of SDRAM at PA=0x20000000
	pTranslationTable[0x200] =
		(uint32_t)finePageTable | //Fine page table base address
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x3;				// This is Fine page table descriptor
	
	//User area: PA from 0x20100000, VA from 0x40000000
	pTranslationTable[0x400] =
		(0x201 << 20) |		// Physical Address
		(1 << 10) |			// Access is free
		(1 << 11) |			// 
		(15 << 5) |			// Domain 15
		(1 << 4) |
		(1 << 3) |			// Cachable
		(1 << 2) |			// Bufferable
		0x2;				// Set as 1 Mbyte section
	
	//Access to physical addresses of user area (only in supervisor mode)
	pTranslationTable[0x201] =
		(0x201 << 20) |		// Physical Address
		(1 << 10) |			// Access in SVC: R/W, in USR: none
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section
	
	//noncachable area of SDRAM
	for(i=0x202;i<0x210;i++)
	{
		pTranslationTable[i] =
			(i << 20) |			// Physical Address
			(1 << 10) |			// Access in SVC: R/W, in USR: none
			(15 << 5) |			// Domain 15
			(1 << 4) |
			(1 << 2) |			// Bufferable
			0x2;				// Set as 1 Mbyte section
	}
	
	//peripherals
	pTranslationTable[0xFFF] =
		(0xFFF << 20) |		// Physical Address
		(1 << 10) |			// Access in SVC: R/W, in USR: none
		(15 << 5) |			// Domain 15
		(1 << 4) |
		0x2;				// Set as 1 Mbyte section*/
	
}


void mmuInit(void)
{
	//initialize the Translation Table
	initTranslationTable(translationTable);
	
	//initialize coarse page table of the 1st MB of SDRAM
	initFinePageTable(finePageTable);
	
	//Set the base address of Translation Table
	AT91F_ARM_WriteTTB((uint32_t)translationTable);
	
	//Program the domain access register
	//domain 15: accesses are checked (client)
	AT91F_ARM_WriteDomain(0x40000000);
	
	//Enable the MMU
	debug_msg("Enabling the MMU...");
	AT91F_EnableMMU();
	debug_txt("ok");	//display message if it works or hangs...
	
	debug_msg("Enabling caches...");
	AT91F_EnableICache();
	AT91F_EnableDCache();
	debug_txt("ok");
	
	
}


//reads and returns CPSR register
inline uint32_t getCpsr(void)
{
	uint32_t res;
	asm volatile("mrs %0,cpsr":"=r"(res));
	return(res);
}


inline uint32_t getPc(void)
{
	register volatile uint32_t pc asm("r15");
	return(pc);
}




void swiHandler(void)
{
	
	//address of SWI instruction is stored in R12
	register volatile uint32_t *swiInstruction asm ("r12");
	uint32_t swiNumber = *swiInstruction & 0x00FFFFFF;
	debug_msg("SWI Exception! SWI number="); debug32_t(swiNumber);
	
	switch(swiNumber)
	{
		case 0x1001: LED_R_TOGGLE; break;
		case 0x1002: LED_G_TOGGLE; break;
		case 0x1003: LED_Y_TOGGLE; break;
		case 0x1004: LED_B_TOGGLE; break;
	}
}


//pass address of bad instruction as an argument
void displayAbortInfo(uint32_t *badInstruction)
{
	debug_msg("Instruction that caused abort was: ");
		debug32_t((uint32_t)*badInstruction);
	debug_txt(", at address: ");
		debug32_t((uint32_t)badInstruction);
	debug_msg("Data Fault Status: "); debug32_t(AT91F_ARM_ReadDataFSR());
	debug_msg("Prefetch Fault Status: "); debug32_t(AT91F_ARM_ReadPrefetchFSR());
	debug_msg("Fault Address Register: "); debug32_t(AT91F_ARM_ReadFAR());
	debug_msg("press any key to continue...");
	debug_waitkey();
}

//handles prefetch abort generated by user program
void userProgramprefetchAbortHandler(void)
{
	register volatile uint32_t *badInstruction asm ("r12");
	uint32_t *temp = (uint32_t*)badInstruction;
	debug_msg("*** User program caused Prefetch Abort!");
	displayAbortInfo(temp);
}

//handles data abort generated by user program
void userProgramDataAbortHandler(void)
{
	register volatile uint32_t *badInstruction asm ("r12");
	uint32_t *temp = (uint32_t*)badInstruction;
	debug_msg("*** User program caused Data Abort!");
	displayAbortInfo(temp);
}

//handles data abort generated by supervisor program
void supervisorDataAbortHandler(void)
{
	register volatile uint32_t *badInstruction asm ("r12");
	uint32_t *temp = (uint32_t*)badInstruction;
	debug_msg("*** Supervisor program caused Data Abort!");
	displayAbortInfo(temp);
}

//handles prefetch abort generated by supervisor program
void supervisorPrefetchAbortHandler(void)
{
	register volatile uint32_t *badInstruction asm ("r12");
	uint32_t *temp = (uint32_t*)badInstruction;
	debug_msg("*** Supervisor program caused Prefetch Abort!");
	displayAbortInfo(temp);
}



void init(void)
{
	// enable reset-key on board 
	AT91C_BASE_RSTC->RSTC_RMR = (0xA5000000 | AT91C_RSTC_URSTEN);
	
	// enable peripheral clock for PIO controller
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA) | (1<<AT91C_ID_PIOB) | (1<<AT91C_ID_PIOC);
	
	LEDS_INIT;
	debug_init_default();
	
	debug_msg("AT91SAM9260 MMU tester");
	
	mmuInit();
	
	fsInit();
	
}



void runProgramFromSD(const char *fileName)
{
//	debug_msg("running program: "); debug_txt((char*)fileName);

	FRESULT res;
	res = f_open(&file,fileName,FA_READ);
	if(res != FR_OK)
	{
		debug_msg("could not open the file, result="); debug8_t(res);
		return;
	}
	
	debug_msg("disabling ICache...");
	AT91F_DisableICache();
	debug_txt("OK");
	
	debug_msg("resetting ICache...");
	AT91F_ARM_InvalidateICache();
	debug_txt("OK");
	
	debug_msg("disabling DCache...");
	AT91F_DisableDCache();
	debug_txt("OK");
	
	debug_msg("resetting DCache...");
	AT91F_CleanDCache();
	AT91F_ARM_DrainWriteBuffer();
	AT91F_ARM_InvalidateDCache();
	debug_txt("OK");
	
	uint32_t *programArea = (uint32_t*)0x40000000;
	
	UINT bytesRead=0;
	res = f_read(&file,(void*)programArea,0xFFFF,&bytesRead);
	if(res != FR_OK)
	{
		debug_msg("program copying error, result="); debug8_t(res);
		return;
	}
	debug_value("bytes read: ", bytesRead);
	
	f_close(&file);
	
	//debug_msg("CPSR="); debug32_t(getCpsr());
	
	debug_msg("reenabling Cache...");
	AT91F_EnableICache();
	AT91F_EnableDCache();
	debug_txt("OK");
	
	debug_msg("starting user program");
	
	asm volatile(
	"stmfd sp!, {r0,r1} \n\t"		//save r0 and r1 on supervisor's stack
	"ldr r0,=0x40000000 \n\t"		//set address of user program to jump in r0
	"mov r1,pc \n\t"				//compute go-back address and store it in r1
	"add r1,r1,#12 \n\t"			//address to go back is current PC+12 bytes
	"stmfd sp!, {r1-r12,lr} \n\t"	//save registers to supervisor's stack
	"msr cpsr,0x10 \n\t"			//switch to user mode
	"mov pc, r0 \n\t"				//jump to address of user program
	
	//this is the first instruction after go back
	"msr CPSR_c, 0x13 \n\t"			//clear I and F flags (SWI exception sets I)
	"ldmia sp!, {r0,r1} \n\t"		//restore other registers
	);
	
	debug_msg("back in main program");
}



int main( void )
{
	init();
	
	debug_msg("entering main loop");
	
	for(;;)
	{
		uint8_t key = debug_inkey();
		
		switch(key)
		{
			case 't':
			{
				debug_msg("test array");
				
				for(uint32_t i=0;i<TEST_ARRAY_SIZE;i++)
				{
					testArray[i] = ~i;
				}
				
				for(uint32_t i=0;i<TEST_ARRAY_SIZE;i++)
				{
					if(testArray[i] != (~i))
					{
						debug_msg("testArray["); debug32_t(i); debug_txt("] = "); debug32_t(testArray[i]);
						debug_txt(" press any key to continue");
						debug_waitkey();
					}
				}
				debug_msg("done");
				break;
			}
			case '1':
				debug_msg("generating SWI 1");
				asm volatile("SWI"::);
				break;
			case '2':
				debug_msg("generating SWI 2");
				asm volatile("SWI 2"::);
				break;
			case '3':
				debug_msg("generating SWI 0xABCD");
				asm volatile("SWI 0xABCD"::);
				break;
			case 'R':
				debug_msg("running a program, enter file name: ");
				char programName[8+1+3+1];
				debug_input_string(programName,8+1+3+1);
				runProgramFromSD(programName);
				break;
			case 'g':
				debug_msg("Get register. Enter address: ");
				uint32_t *addr = (uint32_t*)debug_input_hex(0);
				debug_msg("value read: ");
				debug32_t(*addr);
				break;
		}
		
	}
	return(0);
}

