#include "common.h"
#include "stereo_codec.h"
#include "fat/ff.h"
#include "deck_core_wave.h"

#define WAVE_BUFFER_SIZE_SAMPLES	(16*1024-4)	//15K is quite OK
#define NUMBER_OF_BUFFERS			2
#define BYTES_PER_SAMPLE			2
#define NUMBER_OF_CHANNELS			2

#define SAMPLE_RATE					44100

//RIFF type chunk
typedef struct
{
	uint32_t chunkId;	//"RIFF" text
	uint32_t chunkDataSize;	//total file size - 8
	uint32_t riffType;	//"WAVE" text
}TRIFFHeader;

//Wave File chunks

//Format chunk (fmt)
typedef struct
{
	uint32_t chunkId;	//"fmt " text
	uint32_t chunkDataSize; //usually 0x16 (22 bytes)
	uint16_t compressionCode; //0x0001 for standard PCM format
	uint16_t numberOfChannels; //1 for mono signal, 2 for stereo and so on...
	uint32_t sampleRate; //how many samples per second (i.e. 44100 for audio CD)
	uint32_t averageBytesPerSecond; //how many bytes per second (ie. for stereo audio CD: 44100 16-bit samples gives 2*88200 bytes per second)
	uint16_t blockAlign; //number of bytes occupied by one sample
	uint16_t significantBitsPerSample; //number of bits per one sample in one channel
}TWaveFormatChunk;

//data chunk header, followed by data bytes
typedef struct
{
	uint32_t chunkId;	//"data" text
	uint32_t chunkDataSize; //number of data bytes following
}TWaveDataHeader;

//header structures
static TRIFFHeader			riffHeader;
static TWaveFormatChunk	waveFormatChunk;
static TWaveDataHeader		waveDataHeader;
static uint16_t waveBuffer[NUMBER_OF_BUFFERS][WAVE_BUFFER_SIZE_SAMPLES][NUMBER_OF_CHANNELS]__attribute__((section(".buffers")));
static FIL *pWaveFilePlay __attribute__((section(".buffers")));
static FIL *pWaveFileRec __attribute__((section(".buffers")));
static uint32_t peak[NUMBER_OF_CHANNELS];
//static uint32_t rmsOutput[NUMBER_OF_CHANNELS];


uint32_t deckCoreWaveInit(void)
{
	debug_msg("deckCoreWaveInit...");
	return(OK);
}


uint32_t deckCoreWaveGetCurrentPlayPos(void)
{
	return(pWaveFilePlay->fptr);
}

uint32_t deckCoreWaveGetCurrentRecordPos(void)
{
	return(waveDataHeader.chunkDataSize);
}


uint32_t deckCoreWaveStartRecord(FIL *pFile)
{
	debug_msg("*** entering deckCoreWaveStartRecord");
	pWaveFileRec = pFile;
	
	//wave main header
	riffHeader.chunkId = 0x46464952; //"RIFF" text, magic value
	riffHeader.chunkDataSize =		//size of file - 8 bytes
		sizeof(TRIFFHeader)+sizeof(TWaveFormatChunk)+sizeof(TWaveDataHeader)-8;
	riffHeader.riffType = 0x45564157; //"WAVE" text, magic value
	
	//format chunk
	waveFormatChunk.chunkId = 0x20746D66; //"fmt " (space at end), magic value
	waveFormatChunk.chunkDataSize = 0x16; //size of data in this chunk
	waveFormatChunk.compressionCode = 0x0001; //0x0001 for standard PCM format
	waveFormatChunk.numberOfChannels = NUMBER_OF_CHANNELS;
	waveFormatChunk.sampleRate = SAMPLE_RATE;		//number of samples per second
	waveFormatChunk.averageBytesPerSecond = SAMPLE_RATE * NUMBER_OF_CHANNELS * BYTES_PER_SAMPLE;
	waveFormatChunk.blockAlign = NUMBER_OF_CHANNELS * BYTES_PER_SAMPLE;
	waveFormatChunk.significantBitsPerSample = 8 * BYTES_PER_SAMPLE; //number of bits per one sample in each channel
	
	//data chunk header
	waveDataHeader.chunkId = 0x61746164;
	waveDataHeader.chunkDataSize = 0; //still zero, will be updated during recording
	
	//a variable that stores number of bytes written
	UINT written = 0;
	
	debug_msg("writing headers");
	
	//write headers to file
	if(f_write(pWaveFileRec,&riffHeader,sizeof(TRIFFHeader),&written) != FR_OK) return(FAIL);
	if(f_write(pWaveFileRec,&waveFormatChunk,sizeof(TWaveFormatChunk),&written) != FR_OK) return(FAIL);
	if(f_write(pWaveFileRec,&waveDataHeader,sizeof(TWaveDataHeader),&written) != FR_OK) return(FAIL);
	
	debug_msg("starting RX DMA transfer from codec");
	
	codecStartRx();
	
	uint16_t *pWaveBuffer = &waveBuffer[0][0][0];
	for(uint32_t i=0;i<(NUMBER_OF_CHANNELS*NUMBER_OF_BUFFERS*WAVE_BUFFER_SIZE_SAMPLES);i++)
	{
		*pWaveBuffer++ = 0;
	}
	
	return(OK);
}


uint32_t deckCoreWaveFinishRecording(void)
{
	debug_msg("deckCoreWaveFinishRecording...");
	//a variable that stores number of bytes written
	UINT written = 0;
	
	//set filepointer to start byte
	f_lseek(pWaveFileRec,0);
	
	//write file headers again
	if(f_write(pWaveFileRec,&riffHeader,sizeof(TRIFFHeader),&written)!=FR_OK) debug_msg("header 1 write error");
	if(f_write(pWaveFileRec,&waveFormatChunk,sizeof(TWaveFormatChunk),&written)!=FR_OK)  debug_msg("header 2 write error");
	if(f_write(pWaveFileRec,&waveDataHeader,sizeof(TWaveDataHeader),&written)!=FR_OK)  debug_msg("header 3 write error");
	
	codecStopRx();
	
	debug_msg("deckCoreWaveFinishRecording exit");
	return(OK);
}



uint32_t deckCoreWaveStartPlay(FIL *pFile)
{
	pWaveFilePlay = pFile;	//copy parameter to global
	
	//clear waveBuffer
	uint16_t *pWaveBuffer = &waveBuffer[0][0][0];
	for(uint32_t i=0;i<(NUMBER_OF_CHANNELS*NUMBER_OF_BUFFERS*WAVE_BUFFER_SIZE_SAMPLES);i++)
	{
		*pWaveBuffer++ = 0;
	}
	
	//set file pointer to start of file
	f_lseek(pWaveFilePlay,0);
	
	//read headers
	UINT read;
	FRESULT res;
	res = f_read(pWaveFilePlay,&riffHeader,sizeof(TRIFFHeader),&read);
	res = f_read(pWaveFilePlay,&waveFormatChunk,sizeof(TWaveFormatChunk),&read);
	res = f_read(pWaveFilePlay,&waveDataHeader,sizeof(TWaveDataHeader),&read);
	
	//a good place to analyze header contentz...
	
	//compute and display number of samples
	uint32_t numberOfSamples = waveDataHeader.chunkDataSize / 4;
	debug_value("number of samples = ",numberOfSamples);
	
	//start the DAC
	codecStartTx();
	
	//feed DAC with something on start
	codecTxFeed(&waveBuffer[0][0][0],0x1000);
	
	return(OK);
}


uint32_t deckCoreWaveRestartPlay(void)
{
	//just start the DAC
	codecStartTx();
	return(OK);
}


uint32_t deckCoreWaveStop(void)
{
	//stop the DAC only
	codecStopTx();
	return(OK);
}


uint32_t deckCoreWaveFastForward(void)
{
	const uint32_t BYTES_TO_SKIP = 1024 * 1024;
	if( (pWaveFilePlay->fptr + BYTES_TO_SKIP) > pWaveFilePlay->fsize ) return(FAIL);
	f_lseek(pWaveFilePlay,pWaveFilePlay->fptr + BYTES_TO_SKIP);
	uint16_t *pWave = (uint16_t*)&waveBuffer[0][0][0];
	for(uint32_t i=0;i<(NUMBER_OF_BUFFERS*WAVE_BUFFER_SIZE_SAMPLES*NUMBER_OF_CHANNELS);i++)
		*pWave++ = 0;
	codecTxFeed(&waveBuffer[0][0][0],0x1000);
	return(OK);
}


uint32_t deckCoreWaveRewind(void)
{
	const uint32_t BYTES_TO_REWIND = 1024 * 1024;
	if( BYTES_TO_REWIND > pWaveFilePlay->fptr ) return(FAIL);
	f_lseek(pWaveFilePlay,pWaveFilePlay->fptr - BYTES_TO_REWIND);
	uint16_t *pWave = (uint16_t*)&waveBuffer[0][0][0];
	for(uint32_t i=0;i<(NUMBER_OF_BUFFERS*WAVE_BUFFER_SIZE_SAMPLES*NUMBER_OF_CHANNELS);i++)
		*pWave++ = 0;
	codecTxFeed(&waveBuffer[0][0][0],0x1000);
	return(OK);
}


static void computePeakValues(uint32_t bufIndex)
{
	peak[0] = 0;
	peak[1] = 0;
	for(uint32_t i=0;i<WAVE_BUFFER_SIZE_SAMPLES;i++)
	{
		int16_t sample;
		sample = waveBuffer[bufIndex][i][0];	//get left channel
		if(sample < 0) sample = -sample;	//compute absolute value
		if(peak[0] < sample) peak[0] = sample;
		sample = waveBuffer[bufIndex][i][1];	//get right channel
		if(sample < 0) sample = -sample;	//compute absolute value
		if(peak[1] < sample) peak[1] = sample;
	}
}



uint32_t deckCoreWavePlay(void)
{
	if(codecTxNeedsData())
	{
		UINT bytesRead=0;
		uint32_t bufIndex;
		
		ONE_MORE_SHOT:
		
		bufIndex = codecGetTxBufferIndex();
		
		FRESULT res = f_read(pWaveFilePlay,
			(void*)&waveBuffer[bufIndex][0][0],
			NUMBER_OF_CHANNELS*WAVE_BUFFER_SIZE_SAMPLES*BYTES_PER_SAMPLE,
			&bytesRead);
		
		if(res != FR_OK)
		{
			debug_msg("deckCoreWave, f_read ERROR, res="); debug8_t(res);
		}
		
		if(bytesRead==0)
		{
			debug_value("EOF!!! read bytes ",bytesRead);
			return(DECK_CORE_END_OF_TRACK);
		}
		
		if(codecTxUnderrun())
		{
			debug_msg("ERR deckCoreWave: TX Underrun!");
			codecTxFeed(&waveBuffer[bufIndex][0][0],bytesRead/BYTES_PER_SAMPLE);
			goto ONE_MORE_SHOT;
		}
		else
		{
			//feed the DAC
			codecTxFeed(&waveBuffer[bufIndex][0][0],bytesRead/BYTES_PER_SAMPLE);
		}
		
		computePeakValues(bufIndex);
		
	}
	//else debug_chr('#');
	
	return(DECK_CORE_OK);
}


uint32_t deckCoreWaveRecord(void)
{
	if(codecRxNeedsData())
	{
		UINT bytesWritten=0;
		uint32_t bufIndex;
		bufIndex = codecGetRxBufferIndex();
		FRESULT res = f_write(pWaveFileRec,
			&waveBuffer[bufIndex][0][0],
			NUMBER_OF_CHANNELS*WAVE_BUFFER_SIZE_SAMPLES*BYTES_PER_SAMPLE,
			&bytesWritten);
		
		computePeakValues(bufIndex);
		
		if(res != FR_OK)
		{
			debug_msg("deckCoreWave: f_write error, res="); debug8_t(res);
			return(DECK_CORE_ERROR);
		}
		if(codecRxUnderrun())
		{
			debug_msg("ADC Underrun!");
			codecRxFeed(&waveBuffer[1-bufIndex][0][0],NUMBER_OF_CHANNELS*WAVE_BUFFER_SIZE_SAMPLES);
		}
		codecRxFeed(&waveBuffer[bufIndex][0][0],NUMBER_OF_CHANNELS*WAVE_BUFFER_SIZE_SAMPLES);
		riffHeader.chunkDataSize += bytesWritten;
		waveDataHeader.chunkDataSize += bytesWritten;
		
	}
	
	return(DECK_CORE_OK);
}




void deckCoreGetPeakLevels(uint32_t *peakArray)
{
	peakArray[0] = peak[0];
	peakArray[1] = peak[1];
}





