#include "common.h"
#include "deck_core_mp3.h"
#include "deck_core_wave.h"
#include "stereo_codec.h"
#include "id3.h"
#include "deck_task.h"

static xQueueHandle deckCommandQueue = NULL;
static xQueueHandle deckStatusQueue = NULL;
static xQueueHandle deckInfoQueue = NULL;
static const uint32_t DECK_TASK_QUEUE_LENGTH = 1;
static FIL file __attribute__((section(".buffers")));
static FIL *pCurrentFile = &file;
static T_TrackInfo trackInfo;

enum {
	MODE_IDLE = 0,
	MODE_PLAY_MP3 = 1,
	MODE_PLAY_WAVE = 2,
	MODE_RECORD_WAVE = 3,
};

//static uint32_t mode = MODE_IDLE;

static uint32_t startPlayback(FIL *pFile, const char *fileName)
{
	debug_msg("executing startPlayback command...");
//	uint8_t retval = DECK_STA_OK;
	FRESULT res = f_open(pFile,fileName,FA_READ);
	if(res != FR_OK)
	{
		debug_msg("player_task.c, startPlayback, f_open returned error="); debug8_t(res);
		debug_msg("tried to open fileName: "); debug_txt((char*)fileName);
		return(DECK_STA_ERROR);
	}
	
	char extension[3];
	strcpy(extension,&fileName[strlen(fileName)-3]);
	debug_msg("file extension to play: "); debug_txt(extension);
	
	strlwr(extension);
	
	trackInfo.deckTaskState = DECKTASK_STATE_IDLE;
	trackInfo.totalBytes = pFile->fsize;
	
	if(strcmp(extension,"mp3")==0)	//file with mp3 extension
	{
		trackInfo.deckTaskState = DECKTASK_STATE_PLAY_MP3;
		debug_msg("starting to play MP3 file");
	}
	if(strcmp(extension,"wav")==0)	//file with wav extension
	{
		trackInfo.deckTaskState = DECKTASK_STATE_PLAY_WAVE;
		debug_msg("starting to play wave file");
	}
	
	switch(trackInfo.deckTaskState)
	{
		case DECKTASK_STATE_PLAY_MP3:
		{
			if(deckCoreMp3Start(pFile) != OK)
			{
				debug_msg("player_task.c, startPlayback, playerCoreStart returned FAIL");
				return(DECK_STA_ERROR);
			}
			return(DECK_STA_OK);
		}//case PLAYER_FORMAT_MP3:
		
		case DECKTASK_STATE_PLAY_WAVE:
		{
			if(deckCoreWaveStartPlay(pFile) != OK)
			{
				debug_msg("deck_task.c, startPlayback, deckCoreWaveStart returned bad status");
				return(DECK_STA_ERROR);
			}
			
			return(DECK_STA_OK);
		}//case PLAYER_FORMAT_WAVE:
	}
	
	return(DECK_STA_ERROR);
}


static uint32_t stopCore(FIL *pFile)
{
	debug_msg("deck_task.c, stopCore");
	switch(trackInfo.deckTaskState)
	{
		case DECKTASK_STATE_PLAY_MP3:
			deckCoreMp3Stop();
			break;
		case DECKTASK_STATE_PLAY_WAVE:
			deckCoreWaveStop();
			break;
		case DECKTASK_STATE_RECORD_WAVE:
			deckCoreWaveFinishRecording();
			break;
	}
	trackInfo.deckTaskState = DECKTASK_STATE_IDLE;
	f_close(pFile);
	return(DECK_STA_OK);
}




static uint32_t startRecord(FIL *pFile, const char *fileName)
{
	debug_msg("executing startRecord command...");
	FRESULT res = f_open(pFile,fileName,FA_WRITE | FA_CREATE_ALWAYS);
	if(res != FR_OK)
	{
		debug_msg("deck_task.c, startRecord, f_open returned error="); debug8_t(res);
		debug_msg("tried to open fileName: "); debug_txt((char*)fileName);
		return(DECK_STA_ERROR);
	}
	
	if(deckCoreWaveStartRecord(pFile) != OK)
	{
		debug_msg("deck_task.c, startRecord, deckCoreWaveStart returned bad status");
		return(DECK_STA_ERROR);
	}
	
	trackInfo.deckTaskState = DECKTASK_STATE_RECORD_WAVE;
	trackInfo.sampleRate = 44100;
	
	return(DECK_STA_OK);
}



static uint32_t fastForward(FIL *pFile)
{
	switch(trackInfo.deckTaskState)
	{
		case DECKTASK_STATE_PLAY_MP3:
			if(deckCoreMp3FastForward() == OK)
			{
				return(DECK_STA_OK);
			}
			else
			{
				return(DECK_STA_ERROR);
			}
			break;
		case DECKTASK_STATE_PLAY_WAVE:
			if(deckCoreWaveFastForward() == OK)
			{
				return(DECK_STA_OK);
			}
			else
			{
				return(DECK_STA_ERROR);
			}
			break;
	}
	return(DECK_STA_OK);
}


static uint32_t rewindBackward(FIL *pFile)
{
	switch(trackInfo.deckTaskState)
	{
		case DECKTASK_STATE_PLAY_MP3:
			if(deckCoreMp3Rewind() == OK)
			{
				return(DECK_STA_OK);
			}
			else
			{
				return(DECK_STA_ERROR);
			}
			break;
		case DECKTASK_STATE_PLAY_WAVE:
			if(deckCoreWaveRewind() == OK)
			{
				return(DECK_STA_OK);
			}
			else
			{
				return(DECK_STA_ERROR);
			}
			break;
	}
	return(DECK_STA_OK);
}





static inline void sendStatus(T_DeckStatus *pStatus)
{
	if( xQueueSend(deckStatusQueue,pStatus,1000) != pdTRUE )
	{
		debug_msg("deck_task.c, sendStatus: tried to send error/EOT status, but timeout occured! status=");
		debug32_t(pStatus->status);
	}
}


static void executeCommand(T_DeckCommand *pCommand, T_DeckStatus *pStatus, FIL *pFile)
{
	switch(pCommand->command)
	{
		case DECK_CMD_START_PLAYBACK: pStatus->status = startPlayback(pFile,pCommand->fileName); break;
		case DECK_CMD_STOP_PLAYBACK: pStatus->status = stopCore(pFile); break;
		case DECK_CMD_START_RECORD: pStatus->status = startRecord(pFile,pCommand->fileName); break;
		case DECK_CMD_STOP_RECORD: pStatus->status = stopCore(pFile); break;
		case DECK_CMD_FASTFORWARD: pStatus->status = fastForward(pFile); break;
		case DECK_CMD_REWIND: pStatus->status = rewindBackward(pFile); break;
		default:
			pStatus->status = DECK_STA_ERROR;
			pStatus->parameter = 0;
			break;
	}
}


static void deckTask(void *pvParameters)
{
	static T_DeckStatus statusToSend;
	static T_DeckCommand commandReceived __attribute__((section(".buffers")));
	
	debug_msg("deckTask start!");
	vTaskDelay(300);
	debug_msg("deckTask says hello!");
	codecInit();
	debug_msg("Stereo DAC initialized");
	
	while(1)
	{
		uint8_t status = DECK_STA_OK;
		//codecSetPlaybackVolume(0xAC);
		//vTaskDelay(100);
		switch(trackInfo.deckTaskState)
		{
			case DECKTASK_STATE_PLAY_MP3:
				status = deckCoreMp3();
				trackInfo.currentByte = deckCoreMp3GetCurrentPos();
				break;
			case DECKTASK_STATE_PLAY_WAVE:
				status = deckCoreWavePlay();
				trackInfo.currentByte = deckCoreWaveGetCurrentPlayPos();
				break;
			case DECKTASK_STATE_RECORD_WAVE:
				status = deckCoreWaveRecord();
				trackInfo.currentByte = deckCoreWaveGetCurrentRecordPos();
				break;
			case DECKTASK_STATE_IDLE:
				vTaskDelay(100);
				trackInfo.currentByte = 0;
				status = DECK_CORE_OK;
		}
		
		//hanle errorz
		if(status != DECK_CORE_OK)
		{
			debug_value("playerCore returned error = ",status);
			switch(status)
			{
				case DECK_CORE_END_OF_TRACK:
					statusToSend.status = DECK_STA_END_OF_TRACK;
					debug_msg("end of track");
					stopCore(pCurrentFile);
					break;
				case DECK_CORE_DECODE_ERROR:
					statusToSend.status = DECK_STA_ERROR;
					break;
				default:
					statusToSend.status = DECK_STA_ERROR;
					break;
			}
			
			//send status to queue
			sendStatus(&statusToSend);
		}
		
		xQueueSend(deckInfoQueue,&trackInfo,0);
		
		if(xQueueReceive(deckCommandQueue,&commandReceived,0)==pdTRUE)
		{
			//execute command
			executeCommand(&commandReceived,&statusToSend,pCurrentFile);
			
			//send status to queue
			sendStatus(&statusToSend);
		}
		
		//vTaskDelay(1);
	}
}


uint8_t deckInit(void)
{
	deckCommandQueue = xQueueCreate(DECK_TASK_QUEUE_LENGTH, sizeof(T_DeckCommand));
	if( deckCommandQueue == NULL )
	{
		debug_msg("deckInit: player command queue not created!");
		return(FAIL);
	}
	deckStatusQueue = xQueueCreate(DECK_TASK_QUEUE_LENGTH, sizeof(T_DeckStatus));
	if( deckStatusQueue == NULL )
	{
		debug_msg("deckInit: player status queue not created!");
		return(FAIL);
	}
	deckInfoQueue = xQueueCreate(DECK_TASK_QUEUE_LENGTH, sizeof(T_TrackInfo));
	if( deckInfoQueue == NULL )
	{
		debug_msg("deckInit: player track info queue not created!");
		return(FAIL);
	}
	if( deckCoreMp3Init(NULL) == FAIL)
	{
		debug_msg("deckInit: MP3 core init FAILED");
		return(FAIL);
	}
	if( deckCoreWaveInit() == FAIL)
	{
		debug_msg("deckInit: Wave core init FAILED");
		return(FAIL);
	}
	if( xTaskCreate(deckTask,NULL,STACK_SUPER_LARGE,NULL,PRIORITY_REALTIME,NULL) != pdPASS )
	{
		debug_msg("deckInit: deckTask not created!");
		return(FAIL);
	}
	return(OK);
}


void deckPassQueueHandlers(xQueueHandle* pCommandQueue, xQueueHandle* pStatusQueue, xQueueHandle* pTrackInfoQueue)
{
	if(deckCommandQueue == NULL) debug_msg("WARNING: deckCommandQueue: passing NULL queue!");
	if(deckStatusQueue == NULL) debug_msg("WARNING: deckStatusQueue: passing NULL queue!");
	if(pTrackInfoQueue == NULL) debug_msg("WARNING: pTrackInfoQueue: passing NULL queue!");
	*pCommandQueue = deckCommandQueue;
	*pStatusQueue = deckStatusQueue;
	*pTrackInfoQueue = deckInfoQueue;
}



