#include "common.h"
//#include "deck_core_wave.h"
#include "stereo_codec.h"
#include "stereo_codec_ctrl.h"

#define WAVE_BUFFER_SIZE_SAMPLES	(16*1024)	//15K is quite OK
#define NUMBER_OF_BUFFERS			2
#define NUMBER_OF_CHANNELS			2
#define BYTES_PER_SAMPLE			2
#define LEFT_CHANNEL				0
#define RIGHT_CHANNEL				1

static int16_t adcBuffer[NUMBER_OF_BUFFERS][WAVE_BUFFER_SIZE_SAMPLES][NUMBER_OF_CHANNELS]__attribute__((section(".buffers")));
static int16_t dacBuffer[NUMBER_OF_BUFFERS][WAVE_BUFFER_SIZE_SAMPLES][NUMBER_OF_CHANNELS]__attribute__((section(".buffers")));

enum {
	EFFECT_NONE = 0,
	EFFECT_FIR_LOPASS_20 = 1,
	EFFECT_FIR_LOPASS_40 = 2,
	EFFECT_FIR_HIPASS = 3,
	EFFECT_FIR_BANDPASS = 4,
	EFFECT_FIR_BANDSTOP = 5,
	EFFECT_ECHO = 6,
	NUMBER_OF_EFFECTS = 6,
};

static uint32_t effect = EFFECT_NONE;
static uint32_t echoDelay = 8;
static float echoAttenaution = 0.5;


#define DEFAULT_FIR_ORDER		20


const int16_t FIR_LOPASS_20[DEFAULT_FIR_ORDER+1] =
{
227,
296,
492,
799,
1192,
1632,
2076,
2480,
2803,
3011,
3082,
3011,
2803,
2480,
2076,
1632,
1192,
799,
492,
296,
227,
};

const int16_t FIR_LOPASS_40[40+1] =
{
94,
104,
129,
169,
225,
295,
380,
477,
585,
702,
824,
948,
1071,
1190,
1301,
1401,
1488,
1558,
1609,
1641,
1651,
1641,
1609,
1558,
1488,
1401,
1301,
1190,
1071,
948,
824,
702,
585,
477,
380,
295,
225,
169,
129,
104,
94,
};


const int16_t FIR_BANDSTOP[DEFAULT_FIR_ORDER+1] =
{
189,
224,
61,
-312,
-114,
1473,
3251,
1921,
-4341,
-12450,
31470,
-12450,
-4341,
1921,
3251,
1473,
-114,
-312,
61,
224,
189,
};

const int16_t FIR_BANDPASS[DEFAULT_FIR_ORDER+1] =
{
49,
61,
-73,
-621,
-1659,
-2661,
-2598,
-688,
2723,
6059,
7448,
6059,
2723,
-688,
-2598,
-2661,
-1659,
-621,
-73,
61,
49
};

const int16_t FIR_HIPASS[DEFAULT_FIR_ORDER+1] =
{
48,
109,
215,
295,
177,
-340,
-1363,
-2791,
-4306,
-5470,
26837,
-5470,
-4306,
-2791,
-1363,
-340,
177,
295,
215,
109,
48,
};


void init(void)
{
	// enable peripheral clock for PIO controller
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA) | (1<<AT91C_ID_PIOB) | (1<<AT91C_ID_PIOC);
	LEDS_INIT;
	debug_init_default();
	
	debug_msg("AT91SAM9260 Audio Processor");
	
	debug_msg("stereo codec init");
	codecInit();
	debug_msg("wave core init");
	
	debug_msg("clearing buffers...");
	memset(adcBuffer,0,sizeof(adcBuffer));
	memset(dacBuffer,0,sizeof(dacBuffer));
	debug_txt("done");
	
	LED_G_ON;
	LED_R_OFF;
}


static void displayEffectName(uint32_t effectCode)
{
	switch(effectCode)
	{
		default:
		case EFFECT_NONE:
			debug_txt("No effect");
			break;
		case EFFECT_FIR_LOPASS_20:
			debug_txt("Low pass FIR, 20th order");
			break;
		case EFFECT_FIR_LOPASS_40:
			debug_txt("Low pass FIR, 40th order");
			break;
		case EFFECT_FIR_HIPASS:
			debug_txt("High pass FIR");
			break;
		case EFFECT_FIR_BANDPASS:
			debug_txt("Band pass FIR");
			break;
		case EFFECT_FIR_BANDSTOP:
			debug_txt("Band stop FIR");
			break;
		case EFFECT_ECHO:
			debug_txt("Echo");
			break;
	}
}


void copyBuffer(int16_t *dest, int16_t *src, uint16_t len)
{
	while(len--) *dest++ = *src++;
}


//reads a sample from previous or current buffer
static inline int16_t getSample(int16_t *previousBuf, int16_t *currentBuf, int32_t sample)
{
	if(sample<0)
	{
		return(previousBuf[WAVE_BUFFER_SIZE_SAMPLES*NUMBER_OF_CHANNELS + sample]);
	}
	else
	{
		return(currentBuf[sample]);
	}
}

static inline void setSample(int16_t *previousBuf, int16_t *currentBuf, int32_t sample, int16_t value)
{
	if(sample<0)
	{
		previousBuf[WAVE_BUFFER_SIZE_SAMPLES*NUMBER_OF_CHANNELS + sample] = value;
	}
	else
	{
		currentBuf[sample] = value;
	}
}


static void echo(
	int16_t *previousBuf,
	int16_t *currentBuf,
	int16_t *outputBuf,
	uint32_t delaySamples,
	float reflectionVolume,
	uint32_t right
	)
{
	if(right)
	{
		currentBuf++;
		previousBuf++;
		outputBuf++;
	}
	
	for(int32_t sampleNumber=0;sampleNumber<WAVE_BUFFER_SIZE_SAMPLES;sampleNumber++)
	{
		int32_t thisSample = getSample(previousBuf,currentBuf,2*(sampleNumber-delaySamples));
		int32_t echoSample = getSample(previousBuf,currentBuf,2*(sampleNumber));
		echoSample = echoSample + reflectionVolume * thisSample;
		setSample(previousBuf,currentBuf,2*(sampleNumber),echoSample);
		outputBuf[2*sampleNumber] = thisSample;
	}
}



static void fir(int16_t *previousBuf, int16_t *currentBuf, int16_t *outputBuf, const int16_t *coeffs, uint32_t order, uint32_t right)
{
	if(right)
	{
		currentBuf++;
		previousBuf++;
		outputBuf++;
	}
	
	for(uint32_t sampleNumber=0;sampleNumber<WAVE_BUFFER_SIZE_SAMPLES;sampleNumber++)
	{
		int32_t thisSample = 0;
		uint32_t i = order+1;
		while(i--)
		{
			thisSample = thisSample + coeffs[i] * getSample(previousBuf,currentBuf,2*(sampleNumber-i));
		}
		
		thisSample = thisSample >> 16;
		
		outputBuf[2*sampleNumber] = thisSample;
	}
}




void process(int16_t *inputPrevious, int16_t *inputCurrent, int16_t *output)
{
	LED_G_OFF;
	LED_R_ON;
	
	switch(effect)
	{
		case EFFECT_NONE:
		default:
			memcpy(output,inputCurrent,WAVE_BUFFER_SIZE_SAMPLES*NUMBER_OF_CHANNELS*BYTES_PER_SAMPLE);
			break;
		case EFFECT_FIR_LOPASS_20:
			fir(inputPrevious,inputCurrent,output,FIR_LOPASS_20,DEFAULT_FIR_ORDER,LEFT_CHANNEL);
			fir(inputPrevious,inputCurrent,output,FIR_LOPASS_20,DEFAULT_FIR_ORDER,RIGHT_CHANNEL);
			break;
		case EFFECT_FIR_LOPASS_40:
			fir(inputPrevious,inputCurrent,output,FIR_LOPASS_40,40,LEFT_CHANNEL);
			fir(inputPrevious,inputCurrent,output,FIR_LOPASS_40,40,RIGHT_CHANNEL);
			break;
		case EFFECT_FIR_HIPASS:
			fir(inputPrevious,inputCurrent,output,FIR_HIPASS,DEFAULT_FIR_ORDER,LEFT_CHANNEL);
			fir(inputPrevious,inputCurrent,output,FIR_HIPASS,DEFAULT_FIR_ORDER,RIGHT_CHANNEL);
			break;
		case EFFECT_FIR_BANDPASS:
			fir(inputPrevious,inputCurrent,output,FIR_BANDPASS,DEFAULT_FIR_ORDER,LEFT_CHANNEL);
			fir(inputPrevious,inputCurrent,output,FIR_BANDPASS,DEFAULT_FIR_ORDER,RIGHT_CHANNEL);
			break;
		case EFFECT_FIR_BANDSTOP:
			fir(inputPrevious,inputCurrent,output,FIR_BANDSTOP,DEFAULT_FIR_ORDER,LEFT_CHANNEL);
			fir(inputPrevious,inputCurrent,output,FIR_BANDSTOP,DEFAULT_FIR_ORDER,RIGHT_CHANNEL);
			break;
		case EFFECT_ECHO:
			echo(inputPrevious,inputCurrent,output,echoDelay*1024,echoAttenaution,LEFT_CHANNEL);
			echo(inputPrevious,inputCurrent,output,echoDelay*1024,echoAttenaution,RIGHT_CHANNEL);
			break;
	}
	
	LED_R_OFF;
	LED_G_ON;
}



int main( void )
{
	init();
	
	debug_msg("entering main loop");
	static uint32_t playbackVolume = 0x50;	//step 0.25 dB
	static int32_t recordVolume = 0x00;		//step 0.50 dB
	
	codecSetRecordVolume(recordVolume);
	codecSetPlaybackVolume(playbackVolume);
	
	codecStartRx();
	codecStartTx();
	
	for(;;)
	{
		if(codecRxNeedsData())
		{
			if(codecTxUnderrun())
			{
				codecTxFeed((void*)&dacBuffer[0][0][0],WAVE_BUFFER_SIZE_SAMPLES*NUMBER_OF_CHANNELS);
				debug_msg("tx underrun");
			}
			if(codecRxUnderrun())
			{
				codecRxFeed((void*)&dacBuffer[0][0][0],WAVE_BUFFER_SIZE_SAMPLES*NUMBER_OF_CHANNELS);
				debug_msg("rx overflow");
			}
			uint32_t bufIndex;
			bufIndex = codecGetRxBufferIndex();
			process((void*)&adcBuffer[1-bufIndex][0][0],(void*)&adcBuffer[bufIndex][0][0],(void*)&dacBuffer[bufIndex][0][0]);
			codecTxFeed((void*)&dacBuffer[bufIndex][0][0],WAVE_BUFFER_SIZE_SAMPLES*NUMBER_OF_CHANNELS);
			codecRxFeed((void*)&adcBuffer[bufIndex][0][0],WAVE_BUFFER_SIZE_SAMPLES*NUMBER_OF_CHANNELS);
		}
		
		
		uint8_t key = debug_inkey();
		switch(key)
		{
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				effect = key - '0';
				debug_msg("effect set: "); displayEffectName(effect);
				break;
			case '(':
				if(echoDelay < 12) echoDelay++;
				debug_msg("echo delay: "); debug8_t(echoDelay); debug_txt(" Ksamples");
				break;
			case ')':
				if(echoDelay) echoDelay--;
				debug_msg("echo delay: "); debug8_t(echoDelay); debug_txt(" Ksamples");
				break;
			case '<':
				if(echoAttenaution < 0.9) echoAttenaution+=0.1;
				debug_msg("echo attenaution: "); debug_word(echoAttenaution*100); debug_txt(" %");
				break;
			case '>':
				if(echoAttenaution > 0.1) echoAttenaution-=0.1;
				debug_msg("echo attenaution: "); debug_word(echoAttenaution*100); debug_txt(" %");
				break;
			case '[':	//playback volume down
				if(playbackVolume < 0xFC) playbackVolume++;
				codecSetPlaybackVolume(playbackVolume);
				break;
			case ']':	//playback volume up
			{
				if(playbackVolume > 0x00) playbackVolume--;
				codecSetPlaybackVolume(playbackVolume);
				break;
			}
			case '{':	//rec volume down
				if(recordVolume > -128) recordVolume--;
				codecSetRecordVolume(recordVolume);
				break;
			case '}':	//rec volume up
				if(recordVolume < 0x30) recordVolume++;
				codecSetRecordVolume(recordVolume);
				break;
		}//switch(key)
		//LED_B_TOGGLE;
		delay_ms(10);
		
		
		
	}
	return(0);
}

