#include "common.h"
#include "sd.h"
#include "deck_core_wave.h"
#include "stereo_codec.h"
#include "stereo_codec_ctrl.h"

#define FILE_NAME_LEN		(_MAX_LFN+1)

static char fileName[FILE_NAME_LEN] __attribute__((section(".buffers")));
static FATFS fs __attribute__((section(".buffers")));
static FIL file __attribute__((section(".buffers")));

enum {
	MODE_IDLE,
	MODE_PLAY,
	MODE_RECORD,
};

static uint32_t mode = MODE_IDLE;

void init(void)
{
	// enable peripheral clock for PIO controller
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA) | (1<<AT91C_ID_PIOB) | (1<<AT91C_ID_PIOC);
	LEDS_INIT;
	debug_init_default();
	
	debug_msg("Digital Deck");
	
	debug_msg("initializing filesystem...");
	
	FRESULT res = f_mount(0,&fs);
	debug_txt("result: "); debug8_t(res);
	
	debug_msg("stereo codec init");
	codecInit();
	debug_msg("wave core init");
	deckCoreWaveInit();
}






int main( void )
{
	init();
	
	debug_msg("entering main loop");
	uint32_t status;
	static uint32_t playbackVolume = 0x50;	//step 0.25 dB
	static int32_t recordVolume = 0x00;		//step 0.50 dB
	static uint32_t bass = 0;
	static uint32_t treble = 0;
	
	codecSetRecordVolume(recordVolume);
	codecSetPlaybackVolume(playbackVolume);
	
	for(;;)
	{
		uint8_t key = debug_inkey();
		switch(key)
		{
			case 'v':
			{
				uint32_t peak[2];
				deckCoreGetPeakLevels(peak);
				debug_msg("peak value L="); debug32_t(peak[0]); debug_txt(", R="); debug32_t(peak[1]);
				break;
			}
			case '[':	//playback volume down
			{
				if(playbackVolume < 0xFC) playbackVolume++;
				codecSetPlaybackVolume(playbackVolume);
				break;
			}
			case ']':	//playback volume up
			{
				if(playbackVolume > 0x00) playbackVolume--;
				codecSetPlaybackVolume(playbackVolume);
				break;
			}
			case '{':	//rec volume down
				if(recordVolume > -128) recordVolume--;
				codecSetRecordVolume(recordVolume);
				break;
			case '}':	//rec volume up
				if(recordVolume < 0x30) recordVolume++;
				codecSetRecordVolume(recordVolume);
				break;
			case '<':	//bass down
				if(bass > 0) bass--;
				codecSetBass(bass);
				break;
			case '>':	//bass up
				if(bass < 0xF) bass++;
				codecSetBass(bass);
				break;
			case '-':	//treble down
				if(treble > 0) treble--;
				codecSetTreble(treble);
				break;
			case '+':
				if(treble < 0x3) treble++;
				codecSetTreble(treble);
				break;
			case 'p':
			{
				debug_msg("play wave, enter file name: ");
				debug_input_string(fileName,FILE_NAME_LEN);
				FRESULT res = f_open(&file,fileName,FA_READ);
				if(res != FR_OK)
				{
					debug_msg("could not open wave file, res=");
					debug8_t(res);
					break;
				}
				deckCoreWaveStartPlay(&file);
				mode = MODE_PLAY;
				break;
			}
			case 'r':
			{
				debug_msg("record wave, enter file name: ");
				debug_input_string(fileName,FILE_NAME_LEN);
				FRESULT res = f_open(&file,fileName,FA_WRITE | FA_CREATE_ALWAYS);
				if(res != FR_OK)
				{
					debug_msg("could not open wave file, res=");
					debug8_t(res);
					break;
				}
				deckCoreWaveStartRecord(&file);
				mode = MODE_RECORD;
				break;
			}
			case 's':
				debug_msg("stop");
				switch(mode)
				{
					case MODE_PLAY:
						debug_msg("stopping playback");
						deckCoreWaveStop();
						break;
					case MODE_RECORD:
						debug_msg("stopping record");
						deckCoreWaveFinishRecording();
						break;
				}
				f_close(&file);
				mode = MODE_IDLE;
				break;
		}//switch(key)
		LED_B_TOGGLE;
		delay_ms(50);
		
		switch(mode)
		{
			case MODE_IDLE:	//do nothing
				status = DECK_CORE_OK;
				break;
			case MODE_RECORD:
				status = deckCoreWaveRecord();
				if(status == DECK_CORE_ERROR)
				{
					debug_msg("error during deckCoreWaveRecord");
					deckCoreWaveFinishRecording();
					mode = MODE_IDLE;
					debug_msg("back in MODE_IDLE");
				}
				break;
			case MODE_PLAY:
				status = deckCoreWavePlay();
				if(status==DECK_CORE_END_OF_TRACK)
				{
					debug_msg("play end of track condition");
					deckCoreWaveStop();
					f_close(&file);
					mode = MODE_IDLE;
				}
				break;
		}
		
	}
	return(0);
}

