#include "common.h"
#include "stereo_codec_ctrl.h"
#include "stereo_codec.h"


//DAC Config: 384fs
//FTK = 1379555.556 Hz
//FSYS = MCK/6 = 16554666.67 Hz
//FS = 43111.11111
//FSerr = ~2%
//F_tf = 43.1200kHz (measurement)


#define CODEC_PIO			AT91C_BASE_PIOB
#define CODEC_HW_MASK		(AT91C_PB16_TK0 | AT91C_PB17_TF0 | AT91C_PB18_TD0 | AT91C_PB19_RD0)
#define SSC_ABSR			PIO_ASR	//PIO_ASR or PIO_BSR


#define NUMBER_OF_CHANNELS	2
#define SAMPLE_BIT_RES		16


//timer that generates clock signal for CODEC
#define TIMER				AT91C_BASE_TC3
#define TIMER_ID			AT91C_ID_TC3
#define TIOB				0
#define TIO_PIO				AT91C_BASE_PIOB
#define TIO_PIN				AT91C_PB0_TIOA3
#define TIO_ABSR			PIO_BSR

static uint32_t activeBufferRx;
static uint32_t activeBufferTx;



static void codecStartClocks(void)
{
	//Initialize SysClk of CODEC
	
	TIO_PIO->PIO_PDR = TIO_PIN;
	TIO_PIO->TIO_ABSR = TIO_PIN;
	
	//enable the clock of the TIMER peripheral
	AT91C_BASE_PMC->PMC_PCER = 1<<TIMER_ID;
	
	const uint32_t SELECTED_TC_CLKS = 0;
	
	#if TIOB
		//set capture mode register of the timer
		TIMER->TC_CMR = AT91C_TC_WAVE	//bit WAVE=1 => Waveform Mode
			|SELECTED_TC_CLKS			//selected Timer clock freq.
			|AT91C_TC_BCPB_CLEAR		//RB Compare Effect on TIOB
			|AT91C_TC_BCPC_SET			//RC Compare Effect on TIOB
			|AT91C_TC_EEVT_XC0
			|AT91C_TC_WAVESEL_UP_AUTO;	//Waveform Selection: UP mode with automatic trigger on RC compare
		
		//write full-count value into RC
		TIMER->TC_RC = 3;
		
		//write 0.5 of full value to RB
		TIMER->TC_RB = 1;
	#else
		//set capture mode register of the timer
		TIMER->TC_CMR = AT91C_TC_WAVE	//bit WAVE=1 => Waveform Mode
			|SELECTED_TC_CLKS			//selected Timer clock freq.
			|AT91C_TC_ACPA_CLEAR		//RA Compare Effect on TIOA
			|AT91C_TC_ACPC_SET			//RC Compare Effect on TIOA
			|AT91C_TC_WAVESEL_UP_AUTO;	//Waveform Selection: UP mode with automatic trigger on RC compare
		
		//write full-count value into RC
		TIMER->TC_RC = 3;
		
		//write 0.5 of full-count to RA
		TIMER->TC_RA = 1;
	#endif
	
	//enable timer's clock
	TIMER->TC_CCR = AT91C_TC_CLKEN;
	
	//start timer
	TIMER->TC_CCR = AT91C_TC_SWTRG ;
	
	debug_msg("codec clock: ON");
}


void codecStopClocks(void)
{
	AT91C_BASE_PMC->PMC_PCDR = 1<<TIMER_ID;
	debug_msg("codec clock: OFF");
}





static void codecSscConfig(void)
{
	AT91C_BASE_PMC->PMC_PCER = (1<<SSC_ID);
	
	CODEC_PIO->PIO_ODR = CODEC_HW_MASK;
	CODEC_PIO->PIO_PER = CODEC_HW_MASK;
	
	SSC->SSC_CR = AT91C_SSC_SWRST;
	
	SSC->SSC_CMR = 36;	//OK
	
	//prepare masks for TCMR/RCMR registers
	const uint32_t TCMR_PERIOD_MASK = (((SAMPLE_BIT_RES * NUMBER_OF_CHANNELS)/2)-1) << 24;
	const uint32_t TCMR_STTDLY_MASK = 1 << 16;
	const uint32_t RCMR_PERIOD_MASK = TCMR_PERIOD_MASK;
	const uint32_t RCMR_STTDLY_MASK = TCMR_STTDLY_MASK;
	
	//configure TX channel
	
	SSC->SSC_TCMR =
		TCMR_PERIOD_MASK
		| TCMR_STTDLY_MASK
		| AT91C_SSC_START_FALL_RF
		| AT91C_SSC_CKO_CONTINOUS
		| AT91C_SSC_CKS_DIV;
	
	SSC->SSC_TFMR =
		AT91C_SSC_FSOS_NEGATIVE
		| (SAMPLE_BIT_RES-1)<<16
		| ((NUMBER_OF_CHANNELS-1)<<8)
		| AT91C_SSC_MSBF
		| (SAMPLE_BIT_RES-1);
	
	//configure RX channel
	
	SSC->SSC_RCMR =
		RCMR_PERIOD_MASK |
		RCMR_STTDLY_MASK |
		AT91C_SSC_START_TX |
		AT91C_SSC_CKI |
		AT91C_SSC_CKS_TK;
	
	SSC->SSC_RFMR =
		((SAMPLE_BIT_RES-1)<<16)  |
		((NUMBER_OF_CHANNELS-1)<<8) |
		AT91C_SSC_MSBF |
		(SAMPLE_BIT_RES-1);
	
	SSC->SSC_CR = AT91C_SSC_RXEN | AT91C_SSC_TXEN;
	
	CODEC_PIO->PIO_PDR = CODEC_HW_MASK;
	CODEC_PIO->SSC_ABSR = CODEC_HW_MASK;
}


void codecInit(void)
{
	codecSscConfig();
	codecStartClocks();
	codecCtrlReset();
	codecStopClocks();
	codecCtrlInit();
}


void codecStartTx(void)
{
	//start DMA TX transfer
	SSC->SSC_PTCR = AT91C_PDC_TXTEN;
	
	activeBufferTx = 0;
	
}


void codecStartRx(void)
{
	//start DMA RX transfer
	SSC->SSC_PTCR = AT91C_PDC_RXTEN;
	
	activeBufferRx = 0;
	
}


void codecStopTx(void)
{
	//disable DMA TX transfer
	SSC->SSC_PTCR = AT91C_PDC_TXTDIS;
}


void codecStopRx(void)
{
	//disable DMA RX transfer
	SSC->SSC_PTCR = AT91C_PDC_RXTDIS;
}



//returns index of active buffer
inline uint8_t codecGetTxBufferIndex(void)
{
	return(activeBufferTx);
}


//returns index of active buffer
inline uint8_t codecGetRxBufferIndex(void)
{
	return(activeBufferRx);
}


//returns 1 if user needs to feed codec with new data
//use codecGetWritableBufferIndex to get number of buffer with data
inline uint32_t codecTxNeedsData(void)
{
	if( SSC->SSC_SR & AT91C_SSC_ENDTX ) return(1); else return(0);
}

//returns 1 if user needs to feed codec with new data
//use codecGetWritableBufferIndex to get number of buffer with data
inline uint32_t codecRxNeedsData(void)
{
	if( SSC->SSC_SR & AT91C_SSC_ENDRX ) return(1); else return(0);
}


//returns 1 if both TX buffers are empty
uint32_t codecTxUnderrun(void)
{
	if( SSC->SSC_SR & AT91C_SSC_TXBUFE ) return(1); else return(0);
}


//returns 1 if both RX buffers are empty
uint32_t codecRxUnderrun(void)
{
	if( SSC->SSC_SR & AT91C_SSC_RXBUFF ) return(1); else return(0);
}



//feed the CODEC with a new data
//checks buffer's index with codecGetWritableBufferIndex before feeding the CODEC
void codecTxFeed(int16_t *buffer, uint16_t len)
{
	SSC->SSC_TNPR = (uint32_t)buffer;
	SSC->SSC_TNCR = len;
	
	if(activeBufferTx) activeBufferTx=0; else activeBufferTx=1;
}


//feed the CODEC with a new data
//checks buffer's index with codecGetWritableBufferIndex before feeding the CODEC
void codecRxFeed(int16_t *buffer, uint16_t len)
{
	SSC->SSC_RNPR = (uint32_t)buffer;
	SSC->SSC_RNCR = len;
	
	if(activeBufferRx) activeBufferRx=0; else activeBufferRx=1;
}



