#include "common.h"
#include "stereo_codec_ctrl.h"

//mcu dependent defs:
#define CODEC_I2C_PIO				AT91C_BASE_PIOB
#define CODEC_I2C_SDA_MASK			(1<<1)
#define CODEC_I2C_SCL_MASK			(1<<3)

#define CODEC_I2C_MASK				(CODEC_I2C_SDA_MASK|CODEC_I2C_SCL_MASK)

#define CODEC_I2C_ACK				1
#define CODEC_I2C_NACK				0

#define CODEC_RESET_MASK			AT91C_PIO_PB2
#define CODEC_RESET_PIO				AT91C_BASE_PIOB


//pin defs
#define CODEC_I2C_PIN_INIT			{CODEC_I2C_PIO->PIO_PER = CODEC_I2C_MASK; CODEC_I2C_PIO->PIO_ODR = CODEC_I2C_MASK; CODEC_I2C_PIO->PIO_CODR = CODEC_I2C_MASK;}
#define CODEC_I2C_SDA_IS_HIGH		((CODEC_I2C_PIO->PIO_PDSR & CODEC_I2C_SDA_MASK) == CODEC_I2C_SDA_MASK)
#define CODEC_I2C_SDA_1				CODEC_I2C_PIO->PIO_ODR = CODEC_I2C_SDA_MASK
#define CODEC_I2C_SDA_0				CODEC_I2C_PIO->PIO_OER = CODEC_I2C_SDA_MASK
#define CODEC_I2C_SCL_1				CODEC_I2C_PIO->PIO_ODR = CODEC_I2C_SCL_MASK
#define CODEC_I2C_SCL_0				CODEC_I2C_PIO->PIO_OER = CODEC_I2C_SCL_MASK

#define DELAY_MS(a)					delay_ms(a)



static void codecI2cDelay(void);
static void codecI2cStart(void);
static void codecI2cStop(void);
static uint8_t codecI2cRead(uint8_t ack);
static uint8_t codecI2cWrite(uint8_t data);
static void codecI2cInit(void);


void codecCtrlReset(void)
{
	debug_msg("codec: RESET");
	CODEC_RESET_PIO->PIO_PER = CODEC_RESET_MASK;
	CODEC_RESET_PIO->PIO_CODR = CODEC_RESET_MASK;
	CODEC_RESET_PIO->PIO_OER = CODEC_RESET_MASK;
	DELAY_MS(1);
	CODEC_RESET_PIO->PIO_SODR = CODEC_RESET_MASK;
	DELAY_MS(1);
	CODEC_RESET_PIO->PIO_CODR = CODEC_RESET_MASK;
	DELAY_MS(1);
}

uint32_t codecCtrlInit(void)
{
	codecI2cInit();
	codecWriteRegister(0x00,0x0F32);	//clock from internal WSPLL, all modules on
	uint16_t reg = codecReadRegister(0x00);
	debug_msg("register 0x00 contents: "); debug16_t(reg);
	codecWriteRegister(0x01,0x0000);
	codecWriteRegister(0x02,0xA55F);
	codecWriteRegister(0x12,0xC000);
	codecWriteRegister(0x13,0x0000);
	codecWriteRegister(0x21,0x0000);
	
	return(OK);
}





uint32_t codecSelectInput(uint32_t input)
{
	uint16_t reg22;
	reg22 = codecReadRegister(0x22);
	switch(input)
	{
		case CODEC_INPUT_LINE:
			debug_msg("rec input select: line");
			reg22 &=~0xF0C;	//disable SEL_LNA and SEL_MIC bitz
			break;
		case CODEC_INPUT_MIC:
			debug_msg("rec input select: mic");
			reg22 |= 0xF0C;	//enable SEL_LNA and SEL_MIC bitz
			break;
		default: return FAIL;
	}
	debug_msg("new reg22 contents: "); debug16_t(reg22);
	codecWriteRegister(0x22,reg22);
	return OK;
}



//accepts volume in range from 0x30 (+24dB) to 0x80 (-inf dB)
//step 0.5 dB
void codecSetRecordVolume(int32_t volumeCode)
{
	if(volumeCode > 0x30) volumeCode = 0x30;
	if(volumeCode < -128) volumeCode = -128;
	volumeCode = volumeCode & 0xFF;
	debug_msg("codec: record volume set to 0x"); debug8_t(volumeCode);
	codecWriteRegister(0x20,volumeCode<<8 | volumeCode);
}

//accepts volume in range from 0x00 (0 dB) to 0xFC (-inf)
//step 0.25dB
void codecSetPlaybackVolume(uint32_t volumeCode)
{
	if(volumeCode > 0xFC) volumeCode = 0xFC;
	debug_msg("codec: master volume set to 0x"); debug8_t(volumeCode);
	codecWriteRegister(0x10,volumeCode<<8 | volumeCode);
}

void codecSetTreble(uint32_t treble)
{
	uint16_t reg12 = codecReadRegister(0x12);	//read
	//modify
	treble = treble & 0x3;
	debug_msg("codec: treble boost set to 0x"); debug8_t(treble);
	reg12 &= 0xCFCF;	//clear treble bits
	reg12 |= (treble << 12) | (treble <<4);	//write new treble codes of L and R channel
	codecWriteRegister(0x12,reg12);	//write modified value
}

void codecSetBass(uint32_t bass)
{
	uint16_t reg12 = codecReadRegister(0x12);	//read
	bass = bass & 0xF;	//clear unused bits
	debug_msg("codec: bass boost set to 0x"); debug8_t(bass);
	reg12 &= 0xF0F0;	//clear bass bits
	reg12 |= bass | (bass<<8);	//write new bass bits
	codecWriteRegister(0x12,reg12);	//write modified value
}


uint32_t codecRegisterDump(void)
{
	debug_msg("*** Stereo Codec UDA1380 register dump:");
	for(uint32_t i=0;i<5;i++)
	{
		debug_msg("reg "); debug8_t(i); debug_txt(", value="); debug16_t(codecReadRegister(i));
	}
	for(uint32_t i=0x10;i<0x15;i++)
	{
		debug_msg("reg "); debug8_t(i); debug_txt(", value="); debug16_t(codecReadRegister(i));
	}
	for(uint32_t i=0x20;i<0x24;i++)
	{
		debug_msg("reg "); debug8_t(i); debug_txt(", value="); debug16_t(codecReadRegister(i));
	}
	uint32_t i=0x18;
	debug_msg("reg "); debug8_t(i); debug_txt(", value="); debug16_t(codecReadRegister(i));
	i=0x28;
	debug_msg("reg "); debug8_t(i); debug_txt(", value="); debug16_t(codecReadRegister(i));
	return(OK);
}


uint32_t codecWriteRegister(uint8_t address, uint16_t data)
{
	codecI2cStart();	//generate start condition
	if(codecI2cWrite(0x30)!=CODEC_I2C_ACK)	//write device address
		{debug_msg("codecWriteRegister: FAIL (1)"); return(FAIL);}
	if(codecI2cWrite(address)!=CODEC_I2C_ACK)	//write register address
		{debug_msg("codecWriteRegister: FAIL (2)"); return(FAIL);}
	if(codecI2cWrite(data>>8)!=CODEC_I2C_ACK)	//write most significant part
		{debug_msg("codecWriteRegister: FAIL (3)"); return(FAIL);}
	if(codecI2cWrite(data)!=CODEC_I2C_ACK)	//write least significant part
		{debug_msg("codecWriteRegister: FAIL (4)"); return(FAIL);}
	codecI2cStop();		//generate stop condition
	return(OK);
}


uint16_t codecReadRegister(uint8_t address)
{
	codecI2cStart();	//generate start condition
	if(codecI2cWrite(0x30)!=CODEC_I2C_ACK)	//write device address
		{debug_msg("codecReadRegister: FAIL (1)");}
	if(codecI2cWrite(address)!=CODEC_I2C_ACK)	//write register address
		{debug_msg("codecReadRegister: FAIL (2)");}
	codecI2cStart();	//generate repeated start
	if(codecI2cWrite(0x31)!=CODEC_I2C_ACK)	//write device address
		{debug_msg("codecReadRegister: FAIL (3)");}
	
	uint16_t reg;
	reg = codecI2cRead(CODEC_I2C_ACK);	//read most significant part of reg
	reg = reg << 8;
	reg |= codecI2cRead(CODEC_I2C_NACK);	//read least significant part of reg
	codecI2cStop();
	return(reg);
}


static void codecI2cStart(void)
{
	CODEC_I2C_SDA_1;
	codecI2cDelay();
	CODEC_I2C_SCL_1;
	codecI2cDelay();
	CODEC_I2C_SDA_0;
	codecI2cDelay();
	CODEC_I2C_SCL_0;
	codecI2cDelay();
}


static void codecI2cStop(void)
{
	CODEC_I2C_SDA_0;
	codecI2cDelay();
	CODEC_I2C_SCL_1;
	codecI2cDelay();
	CODEC_I2C_SDA_1;
	codecI2cDelay();
}


static uint8_t codecI2cRead(uint8_t ack)
{
	uint8_t i;
	uint8_t data=0;
	CODEC_I2C_SDA_1;
	for(i=0;i<8;i++)
	{
		CODEC_I2C_SCL_0;
		codecI2cDelay();
		CODEC_I2C_SCL_1;
		codecI2cDelay();
		data = data << 1;
		if(CODEC_I2C_SDA_IS_HIGH) data |= 1;
	}
	CODEC_I2C_SCL_0;
	if(ack==CODEC_I2C_ACK) CODEC_I2C_SDA_0; else CODEC_I2C_SDA_1;
	codecI2cDelay();
	CODEC_I2C_SCL_1;
	codecI2cDelay();
	return(data);
}


static uint8_t codecI2cWrite(uint8_t data)
{
	uint8_t i, ack;
	for(i=0;i<8;i++)
	{
		if(data&0x80) CODEC_I2C_SDA_1; else CODEC_I2C_SDA_0;
		data = data << 1;
		codecI2cDelay();
		CODEC_I2C_SCL_1;
		codecI2cDelay();
		CODEC_I2C_SCL_0;
		codecI2cDelay();
	}
	//ack
	CODEC_I2C_SDA_1;
	codecI2cDelay();
	CODEC_I2C_SCL_1;
	codecI2cDelay();
	if(CODEC_I2C_SDA_IS_HIGH) ack=CODEC_I2C_NACK; else ack=CODEC_I2C_ACK;
	codecI2cDelay();
	CODEC_I2C_SCL_0;
	codecI2cDelay();
	return(ack);
}


static void codecI2cDelay(void)
{
	delay_us(200);
}


static void codecI2cInit(void)
{
	CODEC_I2C_PIN_INIT;
	codecI2cStop();
}



