#include "common.h"
#include "fat/ff.h"
#include "lcd.h"

//max X-Y size in pixels
#define BMP_X_SIZE			640
#define BMP_Y_SIZE			480

typedef struct __attribute__((packed))
{
	uint16_t bfType;		//magic "BM" string
	uint32_t bfSize;		//size of file in bytes
	uint16_t bfReserved1;	//zeros
	uint16_t bfReserved2;	//zeros
	uint32_t bfOffBits;		//at which byte data starts
} TBitmapHeader;

typedef struct __attribute__((packed))
{
	uint32_t biSize;		//size of this structure, usually 40
	uint32_t biWidth;		//width of the image
	uint32_t biHeight;		//height of the image
	uint16_t biPlanes;		//number of planes
	uint16_t biBitCount;	//number of bits per pixel
	uint32_t biCompression;	//type of compression
	uint32_t biSizeImage;	//size of image for compressed images
	uint32_t biXPelsPerMeter; //pixels per meter (not used, zero)
	uint32_t biYPelsPerMeter; //pixels per meter (not used, zero)
	uint32_t biClrUsed;		//number of colors used inbitmap (if zero use biBitCount)
	uint32_t biClrImportant;	//numer of colors important for a bitmap. Use zero - all are important
} TBitmapInfoHeader;

typedef struct __attribute__((packed)){
	uint8_t blue;
	uint8_t green;
	uint8_t red;
}TPixel;


static TBitmapHeader bmpHeader __attribute__((section(".buffers")));
static TBitmapInfoHeader bmpInfoHeader __attribute__((section(".buffers")));
static TPixel row[BMP_X_SIZE] __attribute((section(".buffers")));
static FIL bmpFile __attribute((section(".buffers")));

//converts data in 24-bit pixel structure into RGB565 format (returned value)
static inline uint16_t convert24BitToShadow(TPixel *pPixel)
{
	uint16_t outPixel =
		( ((uint32_t)pPixel->red << 8) & LCD_COLOR_RED) |
		( ((uint32_t)pPixel->green << 3) & LCD_COLOR_GREEN) |
		( ((uint32_t)pPixel->blue >> 3) & LCD_COLOR_BLUE );
	return(outPixel);
}

static FRESULT buildHeaders(FIL *pFile, uint32_t xSize, uint32_t ySize)
{
	FRESULT res;
	UINT written;
	const uint32_t BITS = 24;
	
	bmpHeader.bfType = ((uint16_t)('M')<<8) | 'B';
	bmpHeader.bfSize = sizeof(TBitmapHeader) + sizeof(TBitmapInfoHeader) + xSize * ySize * (BITS / 8);	//size of 2 headerz + size of data field
	bmpHeader.bfReserved1 = 0;
	bmpHeader.bfReserved2 = 0;
	bmpHeader.bfOffBits = sizeof(TBitmapHeader) + sizeof(TBitmapInfoHeader);	//offset. data follows headerz
	
	res = f_write(pFile, &bmpHeader, sizeof(TBitmapHeader), &written);
	if(res != FR_OK) {debug_msg("buildHeaders: f_write (1) error, res="); debug8_t(res); return(res);}
	debug_msg("header 1 write OK, bytes wr: "); debug32_t(written);
	
	bmpInfoHeader.biSize = sizeof(TBitmapInfoHeader);
	bmpInfoHeader.biWidth = xSize;
	bmpInfoHeader.biHeight = ySize;
	bmpInfoHeader.biPlanes = 1;	//number of planes
	bmpInfoHeader.biBitCount = BITS;	//number of bitz/pixel
	bmpInfoHeader.biCompression = 0;
	bmpInfoHeader.biSizeImage = 0;
	bmpInfoHeader.biXPelsPerMeter = 0;
	bmpInfoHeader.biYPelsPerMeter = 0;
	bmpInfoHeader.biClrUsed = 0;
	bmpInfoHeader.biClrImportant = 0;
	
	res = f_write(pFile, &bmpInfoHeader, sizeof(TBitmapInfoHeader), &written);
	if(res != FR_OK) {debug_msg("buildHeaders: f_write (2) error, res="); debug8_t(res); return(res);}
	debug_msg("header 2 write OK, bytes wr: "); debug32_t(written);
	
	return(FR_OK);
}

static FRESULT savePixels(FIL *pFile, TPixel *pixelArray, uint32_t pixelCount)
{
	UINT written;
	FRESULT res = f_write(pFile,pixelArray,pixelCount*sizeof(TPixel),&written);
	if(res != FR_OK) {debug_msg("savePixels error, res="); debug8_t(res);}
	return(FR_OK);
}

static void convertYCbCrToRgb(TPixel *rgbPixelArray, uint32_t *yCbCrPixel, uint32_t NbYCbCrPixels)
{
	uint32_t temp;
	uint8_t Y[2];
	uint8_t Cb;
	uint8_t Cr;
	while(NbYCbCrPixels--)
	{
		temp = *yCbCrPixel;	//copy value to temp
		
		//extract Y Cb and Cr components
		Cb = temp;
		temp = temp >> 8;
		Y[0] = temp;
		temp = temp >> 8;
		Cr = temp;
		temp = temp >> 8;
		Y[1] = temp;
		
		//compute 2 RGB pixels
		for(uint32_t i=0;i<2;i++)
		{
			//R
			temp = 1.164 * (Y[i]-16) + 1.596 * (Cr-128);
			if(temp > 255) {/*debug_msg("clipping (1) from "); debug32_t(temp); debug_txt(" to FF");*/ temp = 255;}
			rgbPixelArray->red = temp;
			//G
			temp = 1.164 * (Y[i]-16) - 0.813 * (Cr-128) - 0.392 * (Cb-128);
			if(temp > 255) {/*debug_msg("clipping (2) from "); debug32_t(temp); debug_txt(" to FF");*/ temp = 255;}
			rgbPixelArray->green = temp;
			//B
			temp = 1.164 * (Y[i]-16) + 2.107 * (Cb - 128);
			if(temp > 255) {/*debug_msg("clipping (3) from "); debug32_t(temp); debug_txt(" to FF");*/ temp = 255;}
			rgbPixelArray->blue = temp;
			
			//move rgb pointer
			rgbPixelArray++;
		}
		
		//move YCbCr pointer
		yCbCrPixel++;
	}
}

//provide image size in pixels, 2 pixels are coded in one pYCbCr array element
uint32_t bmpSaveYCbCr(uint32_t *pYCbCr, uint32_t xSize, uint32_t ySize, const char * fileName)
{
	FRESULT res;
	debug_msg("starting bmpSaveYCbCr...");
	
	//open or create a new BMP file
	res = f_open(&bmpFile,fileName,FA_WRITE | FA_CREATE_ALWAYS);
	if(res != FR_OK) {debug_msg("failed to open BMP file, res="); debug8_t(res); return(FAIL);}
	else
	{debug_msg("file opened OK, name: "); debug_txt((char*)fileName);}
	
	//create BMP headers
	res = buildHeaders(&bmpFile,xSize,ySize);
	if(res != FR_OK) {debug_msg("failed to create BMP headers res="); debug8_t(res); return(FAIL);}
	else
	{debug_msg("headers created OK");}
	
	uint32_t* sourcePtr;
	uint32_t line = ySize;
	uint32_t xSizeInYCbCr = xSize/2;
	sourcePtr = &pYCbCr[(line-1)*xSizeInYCbCr];	//set source image pointer to last line
	
	debug_msg("pYCbCr at address: "); debug32_t((uint32_t)pYCbCr);
	debug_msg("sourcePtr at start: "); debug32_t((uint32_t)sourcePtr);
	
	while(line--)
	{
		convertYCbCrToRgb(row,sourcePtr,xSizeInYCbCr);
		if(savePixels(&bmpFile,row,xSize)!=FR_OK) {debug_msg("bmpSaveYCbCr loop aborted"); break;;}
		sourcePtr -= (xSizeInYCbCr);	//move 1 image line up (backward in source array)
		//debug_msg("sourcePtr for line : "); debug32_t(line); debug_txt(" is: "); debug32_t((uint32_t)sourcePtr);
		//debug_waitkey();
	}
	
	f_close(&bmpFile);
	
	//convertYCbCrToRgb(pYCbCr
	return(OK);
}
