#include "common.h"
#include "i2c.h"
#include "bitmap.h"
#include "camera.h"

#define ISI			AT91C_BASE_HISI
#define ISI_ID		AT91C_ID_HISI

#define RES_PIO			AT91C_BASE_PIOB
#define RES_MASK		AT91C_PIO_PB8
#define CAM_RESET_INIT	{RES_PIO->PIO_OER=RES_MASK; RES_PIO->PIO_SODR = RES_MASK;}
#define CAM_RESET_1		RES_PIO->PIO_SODR = RES_MASK
#define CAM_RESET_0		RES_PIO->PIO_CODR = RES_MASK

#define CAM_PIO			AT91C_BASE_PIOB
#define CAM_HW_MASK		( 0x7FF00000 )
#define CAM_MCK_MASK	AT91C_PIO_PB31
#define CAM_MCK_ABSEL	PIO_ASR

#define REG_GROUP_A		0x00
#define REG_GROUP_B		0x01
#define REG_GROUP_C		0x02
#define REG_GROUP_D		0x03

typedef struct
{
	uint32_t currentBufferAddress;
	uint32_t nextFBDAddress;
} TFrameBufferDescriptor;

//allocate static frame buffer descriptor
static volatile TFrameBufferDescriptor fbd __attribute__((section(".buffers")));
//static volatile TFrameBufferDescriptor *pFbd = (TFrameBufferDescriptor *) &fbd;

static inline void camPreviewSize(uint16_t x, uint16_t y, uint8_t decimation);
static uint8_t camWriteReg(uint8_t reg, uint8_t value);
static uint8_t camReadReg(uint8_t reg);
static inline uint8_t camSelectRegGroup(uint8_t group);


static inline void camReset(void)
{
	CAM_RESET_INIT;
	CAM_RESET_1;
	delay_ms(100);
	CAM_RESET_0;
	delay_ms(10);
	CAM_RESET_1;
	delay_ms(10);
	
}


static inline uint8_t camSelectRegGroup(uint8_t group)
{
	if(group > 3) return(FAIL);
	return(camWriteReg(3,group));
}


static uint8_t camWriteReg(uint8_t reg, uint8_t value)
{
	i2cStart();
	uint8_t ack = OK;
	if(i2cWrite(0xDC)==I2C_NACK) ack=FAIL;
	if(i2cWrite(reg)==I2C_NACK) ack=FAIL;
	if(i2cWrite(value)==I2C_NACK) ack=FAIL;
	i2cStop();
	if(ack==FAIL) debug_msg("ERR Cam reg set FAILED!");
	return(ack);
}


static uint8_t camReadReg(uint8_t reg)
{
	i2cStart();
	uint8_t ack = OK;
	if(i2cWrite(0xDC)==I2C_NACK) ack=FAIL;
	if(i2cWrite(reg)==I2C_NACK) ack=FAIL;
	i2cStart();
	if(i2cWrite(0xDD)==I2C_NACK) ack=FAIL;
	uint8_t ret = i2cRead(I2C_NACK);
	i2cStop();
	if(ack==FAIL) debug_msg("camReadReg FAILED, no ACK");
	return(ret);
}

//returns type of a camera connected
uint32_t camGetChipType(void)
{
	uint32_t type;
	
	camSelectRegGroup(REG_GROUP_D);	//PO2030 should ignore this command
	type = camReadReg(0);
	type = type << 8;
	type |= camReadReg(1);
	
	debug_msg("Camera ID: ");
	debug16_t(type);
	switch(type)
	{
		case CAM_TYPE_PO2030:
			debug_txt(" (PO2030)");
			break;
		case CAM_TYPE_PO6030:
			debug_txt(" (PO6030)");
			break;
		default:
			return(CAM_TYPE_NONE);
	}
 	return(type);
}


void camShowTestScreen(uint32_t number)
{
	camSelectRegGroup(REG_GROUP_B);
	camWriteReg(0x40,number);
}


//camera starts with programmer-friendly default settings
//there's not much to do
uint32_t camInitRegisters(uint32_t type)
{
	switch(type)
	{
		case CAM_TYPE_PO2030:
			//your code for PO2030 non-standard init
			return(OK);
		case CAM_TYPE_PO6030:
			//PO6030 non-standard init
			camSelectRegGroup(REG_GROUP_B);
			//set bit "Active High of Hsync All Lines"
			camWriteReg(0x69, 0x26);
			return(OK);
		default:
			return(FAIL);
	}
}



static uint32_t clockInit(void)
{
	//enable PCK1 clock on Power Management Controller
	AT91C_BASE_PMC->PMC_SCER = AT91C_PMC_PCK1;
	
	//configure PCK1 clock
	AT91C_BASE_PMC->PMC_PCKR[1] = AT91C_PMC_CSS_PLLA_CLK | AT91C_PMC_PRES_CLK_8;
	
	//select PCK1 peripheral (Peripheral A)
	CAM_PIO->PIO_ASR = CAM_MCK_MASK;
	
	//disable PIO on camera's interface pins
	CAM_PIO->PIO_PDR = CAM_MCK_MASK;
	
	return(OK);
}


static uint32_t isiPioInit(void)
{
	//select camera's peripheral (Peripheral B)
	CAM_PIO->PIO_BSR = CAM_HW_MASK;
	
	//disable PIO on camera's interface pins
	CAM_PIO->PIO_PDR = CAM_HW_MASK;
	
	CAM_PIO->PIO_PPUDR = CAM_HW_MASK;
	
	return(OK);
}


static void isiInit(void *previewBuffer, void *imageBuffer, uint32_t camType)
{
	//enable peripheral clock for ISI
	AT91C_BASE_PMC->PMC_PCER = 1<<AT91C_ID_HISI;
	
	fbd.currentBufferAddress = (uint32_t) previewBuffer;
	fbd.nextFBDAddress = (uint32_t) &fbd;
	
	//preview buffer descriptor address
	ISI->ISI_PFBD = (uint32_t) &fbd;
	
	camPreviewSize(CAM_PREVIEW_X_SIZE,CAM_PREVIEW_Y_SIZE,CAM_DECIMATION);
	
	//codec buffer base address - a buffer for the final image
	ISI->ISI_CDBA = (uint32_t) imageBuffer;
	
	//initialize ISI Control registers ISI_CR1 and ISI_CR2
	
	ISI->ISI_CR1 = AT91C_ISI_PIXCLK_POL;		//pixel data clocked on falling edge of PCLK, other bits set to zero
	while((ISI->ISI_SR & AT91C_ISI_DIS));
	
	//prepare masks for image size
	const uint32_t CR1_IM_HSIZE_MASK = (CAM_IMAGE_X_SIZE-1) << 16;
	const uint32_t CR1_IM_VSIZE_MASK = (CAM_IMAGE_Y_SIZE-1) << 0;
	
	//set CR2 registers
	ISI->ISI_CR2 =  CR1_IM_HSIZE_MASK | CR1_IM_VSIZE_MASK;
}


//returns cam type code
uint32_t camInit(void *previewBuffer, void *imageBuffer)
{
	uint32_t camType = CAM_TYPE_NONE;
	debug_msg("camera init...");
	
	debug_msg("fbd at address="); debug32_t((uint32_t)&fbd);
	
	//initialize clock output (master clock for camera)
	clockInit();
	
	//initialize I2C bus
	i2cInit();
	
	//setup PIO controllers to interface camera module
	isiPioInit();
	
	//perform camera chip reset
	camReset();
	
	//recognize camera chip
	camType = camGetChipType();
	
	//initialize camera's internal registers
	camInitRegisters(camType);
	
	//initialize ISI module
	isiInit(previewBuffer, imageBuffer, camType);
	
	debug_msg("cam init done");
	
	//inform main program about camera type
	return(camType);	
}

uint32_t camCaptureStillImage(void)
{
	ISI->ISI_CR1 |= AT91C_ISI_CODEC_ON;
	//debug_msg("PND");
	while( (ISI->ISI_SR & AT91C_ISI_CDC_PND) != 0 );	//wait for end of current frame
	//debug_msg("EMP");
	while( (ISI->ISI_SR & AT91C_ISI_FO_C_EMP) == 0 );	//wait for end of DMA transfer
	debug_msg("Frame capture done");
	ISI->ISI_CR1 &= ~AT91C_ISI_CODEC_ON;
	return(OK);
}

static inline void camPreviewSize(uint16_t x, uint16_t y, uint8_t decimation)
{
	ISI->ISI_PDECF = decimation;
	ISI->ISI_PSIZE = (uint32_t)(x-1)<<16 | (y-1);
}




